"""job_condition

Revision ID: 5a5c49313e52
Revises: 4b3c9d63e89c
Create Date: 2024-02-02 11:17:07.246330

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db  # noqa

# revision identifiers, used by Alembic.
revision = '5a5c49313e52'
down_revision = '4b3c9d63e89c'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('jobs', schema=None) as batch_op:
        batch_op.add_column(sa.Column('if_query_str', sa.String(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('jobs', schema=None) as batch_op:
        batch_op.drop_column('if_query_str')

    # ### end Alembic commands ###



from mindsdb.api.executor.datahub.classes.response import DataHubResponse


class DataNode:
    type = 'meta'

    def __init__(self):
        pass

    def get_type(self):
        return self.type

    def get_tables(self):
        pass

    def get_table_columns(self, tableName, schema_name=None):
        pass

    def query(self, query=None, native_query=None, session=None) -> DataHubResponse:
        return []



from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packet import Packet
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_datum import Datum


class SwitchOutResponse(Packet):
    def setup(self, length=0, count_header=1, body=''):
        length = len(body)

        if length == 0:
            self.password = b''
            return

        self.enc_password = Datum('string<EOF>')  # 0x04
        buffer = body
        buffer = self.enc_password.setFromBuff(buffer)
        self.password = self.enc_password.value



import traceback
import json
import csv
from io import BytesIO, StringIO, IOBase
from pathlib import Path
import codecs
from typing import List

import filetype
import pandas as pd
from charset_normalizer import from_bytes
from langchain_text_splitters import RecursiveCharacterTextSplitter
import fitz  # pymupdf

from mindsdb.utilities import log

logger = log.getLogger(__name__)

DEFAULT_CHUNK_SIZE = 500
DEFAULT_CHUNK_OVERLAP = 250


class FileDetectError(Exception):
    ...


def decode(file_obj: IOBase) -> StringIO:
    file_obj.seek(0)
    byte_str = file_obj.read()
    # Move it to StringIO
    try:
        # Handle Microsoft's BOM "special" UTF-8 encoding
        if byte_str.startswith(codecs.BOM_UTF8):
            data_str = StringIO(byte_str.decode("utf-8-sig"))
        else:
            file_encoding_meta = from_bytes(
                byte_str[: 32 * 1024],
                steps=32,  # Number of steps/block to extract from my_byte_str
                chunk_size=1024,  # Set block size of each extraction)
                explain=False,
            )
            best_meta = file_encoding_meta.best()
            errors = "strict"
            if best_meta is not None:
                encoding = file_encoding_meta.best().encoding

                try:
                    data_str = StringIO(byte_str.decode(encoding, errors))
                except UnicodeDecodeError:
                    encoding = "utf-8"
                    errors = "replace"

                    data_str = StringIO(byte_str.decode(encoding, errors))
            else:
                encoding = "utf-8"
                errors = "replace"

                data_str = StringIO(byte_str.decode(encoding, errors))
    except Exception as e:
        logger.error(traceback.format_exc())
        raise FileDetectError("Could not load into string") from e

    return data_str


class FormatDetector:

    supported_formats = ['parquet', 'csv', 'xlsx', 'pdf', 'json', 'txt']
    multipage_formats = ['xlsx']

    def __init__(
        self,
        path: str = None,
        name: str = None,
        file: IOBase = None
    ):
        """
        File format detector
        One of these arguments has to be passed: `path` or `file`

        :param path: path to the file
        :param name: name of the file
        :param file: file descriptor (via open(...), of BytesIO(...))
        """
        if path is not None:
            file = open(path, 'rb')

        elif file is not None:
            if name is None:
                if hasattr(file, 'name'):
                    path = file.name
                else:
                    path = 'file'
        else:
            raise FileDetectError('Wrong arguments: path or file is required')

        if name is None:
            name = Path(path).name

        self.name = name
        self.file_obj = file
        self.format = None

        self.parameters = {}

    def get_format(self) -> str:
        if self.format is not None:
            return self.format

        format = self.get_format_by_name()
        if format is not None:
            if format not in self.supported_formats:
                raise FileDetectError(f'Not supported format: {format}')

        if format is None and self.file_obj is not None:
            format = self.get_format_by_content()
            self.file_obj.seek(0)

        if format is None:
            raise FileDetectError(f'Unable to detect format: {self.name}')

        self.format = format
        return format

    def get_format_by_name(self):
        extension = Path(self.name).suffix.strip(".").lower()
        if extension == "tsv":
            extension = "csv"
            self.parameters['delimiter'] = '\t'

        return extension or None

    def get_format_by_content(self):
        if self.is_parquet(self.file_obj):
            return "parquet"

        file_type = filetype.guess(self.file_obj)
        if file_type is not None:

            if file_type.mime in {
                "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
                "application/vnd.ms-excel",
            }:
                return 'xlsx'

            if file_type.mime == 'application/pdf':
                return "pdf"

        file_obj = decode(self.file_obj)

        if self.is_json(file_obj):
            return "json"

        if self.is_csv(file_obj):
            return "csv"

    @staticmethod
    def is_json(data_obj: StringIO) -> bool:
        # see if its JSON
        data_obj.seek(0)
        text = data_obj.read(100).strip()
        data_obj.seek(0)
        if len(text) > 0:
            # it looks like a json, then try to parse it
            if text.startswith("{") or text.startswith("["):
                try:
                    json.loads(data_obj.read())
                    return True
                except Exception:
                    return False
        return False

    @classmethod
    def is_csv(cls, data_obj: StringIO) -> bool:
        data_obj.seek(0)
        sample = data_obj.readline()  # trying to get dialect from header
        try:
            data_obj.seek(0)
            csv.Sniffer().sniff(sample)

            # Avoid a false-positive for json files
            if cls.is_json(data_obj):
                return False
            return True
        except Exception:
            return False

    @staticmethod
    def is_parquet(data: IOBase) -> bool:
        # Check first and last 4 bytes equal to PAR1.
        # Refer: https://parquet.apache.org/docs/file-format/
        parquet_sig = b"PAR1"
        data.seek(0, 0)
        start_meta = data.read(4)
        data.seek(-4, 2)
        end_meta = data.read()
        data.seek(0)
        if start_meta == parquet_sig and end_meta == parquet_sig:
            return True
        return False


class FileReader(FormatDetector):

    def _get_fnc(self):
        format = self.get_format()
        func = getattr(self, f'read_{format}', None)
        if func is None:
            raise FileDetectError(f'Unsupported format: {format}')
        return func

    def get_pages(self, **kwargs) -> List[str]:
        """
            Get list of tables in file
        """
        format = self.get_format()
        if format not in self.multipage_formats:
            # only one table
            return ['main']

        func = self._get_fnc()
        self.file_obj.seek(0)

        return [
            name for name, _ in
            func(self.file_obj, only_names=True, **kwargs)
        ]

    def get_contents(self, **kwargs):
        """
            Get all info(pages with content) from file as dict: {tablename, content}
        """
        func = self._get_fnc()
        self.file_obj.seek(0)

        format = self.get_format()
        if format not in self.multipage_formats:
            # only one table
            return {'main': func(self.file_obj, name=self.name, **kwargs)}

        return {
            name: df
            for name, df in
            func(self.file_obj, **kwargs)
        }

    def get_page_content(self, page_name: str = None, **kwargs) -> pd.DataFrame:
        """
            Get content of a single table
        """
        func = self._get_fnc()
        self.file_obj.seek(0)

        format = self.get_format()
        if format not in self.multipage_formats:
            # only one table
            return func(self.file_obj, name=self.name, **kwargs)

        for _, df in func(self.file_obj, name=self.name, page_name=page_name, **kwargs):
            return df

    @staticmethod
    def _get_csv_dialect(buffer, delimiter=None) -> csv.Dialect:
        sample = buffer.readline()  # trying to get dialect from header
        buffer.seek(0)
        try:
            if isinstance(sample, bytes):
                sample = sample.decode()

            if delimiter is not None:
                accepted_csv_delimiters = [delimiter]
            else:
                accepted_csv_delimiters = [",", "\t", ";"]
            try:
                dialect = csv.Sniffer().sniff(
                    sample, delimiters=accepted_csv_delimiters
                )
                dialect.doublequote = (
                    True  # assume that all csvs have " as string escape
                )
            except Exception:
                dialect = csv.reader(sample).dialect
                if dialect.delimiter not in accepted_csv_delimiters:
                    raise Exception(
                        f"CSV delimeter '{dialect.delimiter}' is not supported"
                    )

        except csv.Error:
            dialect = None
        return dialect

    @classmethod
    def read_csv(cls, file_obj: BytesIO, delimiter=None, **kwargs):
        file_obj = decode(file_obj)
        dialect = cls._get_csv_dialect(file_obj, delimiter=delimiter)

        return pd.read_csv(file_obj, sep=dialect.delimiter, index_col=False)

    @staticmethod
    def read_txt(file_obj: BytesIO, name=None, **kwargs):
        file_obj = decode(file_obj)

        try:
            from langchain_core.documents import Document
        except ImportError:
            raise ImportError(
                "To import TXT document please install 'langchain-community':\n"
                "    pip install langchain-community"
            )
        text = file_obj.read()

        metadata = {"source_file": name, "file_format": "txt"}
        documents = [Document(page_content=text, metadata=metadata)]

        text_splitter = RecursiveCharacterTextSplitter(
            chunk_size=DEFAULT_CHUNK_SIZE, chunk_overlap=DEFAULT_CHUNK_OVERLAP
        )

        docs = text_splitter.split_documents(documents)
        return pd.DataFrame(
            [
                {"content": doc.page_content, "metadata": doc.metadata}
                for doc in docs
            ]
        )

    @staticmethod
    def read_pdf(file_obj: BytesIO, name=None, **kwargs):

        with fitz.open(stream=file_obj.read()) as pdf:  # open pdf
            text = chr(12).join([page.get_text() for page in pdf])

        text_splitter = RecursiveCharacterTextSplitter(
            chunk_size=DEFAULT_CHUNK_SIZE, chunk_overlap=DEFAULT_CHUNK_OVERLAP
        )

        split_text = text_splitter.split_text(text)

        return pd.DataFrame(
            {"content": split_text, "metadata": [{"file_format": "pdf", "source_file": name}] * len(split_text)}
        )

    @staticmethod
    def read_json(file_obj: BytesIO, **kwargs):
        file_obj = decode(file_obj)
        file_obj.seek(0)
        json_doc = json.loads(file_obj.read())
        return pd.json_normalize(json_doc, max_level=0)

    @staticmethod
    def read_parquet(file_obj: BytesIO, **kwargs):
        return pd.read_parquet(file_obj)

    @staticmethod
    def read_xlsx(file_obj: BytesIO, page_name=None, only_names=False, **kwargs):
        with pd.ExcelFile(file_obj) as xls:

            if page_name is not None:
                # return specific page
                yield page_name, pd.read_excel(xls, sheet_name=page_name)

            for page_name in xls.sheet_names:

                if only_names:
                    # extract only pages names
                    df = None
                else:
                    df = pd.read_excel(xls, sheet_name=page_name)
                yield page_name, df



import json
import logging
from logging.config import dictConfig

from mindsdb.utilities.config import config as app_config


logging_initialized = False


class JsonFormatter(logging.Formatter):
    def format(self, record):
        record_message = super().format(record)
        log_record = {
            'process_name': record.processName,
            'name': record.name,
            'message': record_message,
            'level': record.levelname,
            'time': record.created
        }
        return json.dumps(log_record)


class ColorFormatter(logging.Formatter):
    green = "\x1b[32;20m"
    default = "\x1b[39;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format = "%(asctime)s %(processName)15s %(levelname)-8s %(name)s: %(message)s"

    FORMATS = {
        logging.DEBUG: logging.Formatter(green + format + reset),
        logging.INFO: logging.Formatter(default + format + reset),
        logging.WARNING: logging.Formatter(yellow + format + reset),
        logging.ERROR: logging.Formatter(red + format + reset),
        logging.CRITICAL: logging.Formatter(bold_red + format + reset),
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        return log_fmt.format(record)


def get_console_handler_config_level() -> int:
    console_handler_config = app_config['logging']['handlers']['console']
    return getattr(logging, console_handler_config["level"])


def get_file_handler_config_level() -> int:
    file_handler_config = app_config['logging']['handlers']['file']
    return getattr(logging, file_handler_config["level"])


def get_mindsdb_log_level() -> int:
    console_handler_config_level = get_console_handler_config_level()
    file_handler_config_level = get_file_handler_config_level()

    return min(console_handler_config_level, file_handler_config_level)


def configure_logging():
    handlers_config = {}
    console_handler_config = app_config['logging']['handlers']['console']
    console_handler_config_level = getattr(logging, console_handler_config["level"])
    if console_handler_config['enabled'] is True:
        handlers_config['console'] = {
            "class": "logging.StreamHandler",
            "formatter": console_handler_config.get('formatter', 'default'),
            "level": console_handler_config_level
        }

    file_handler_config = app_config['logging']['handlers']['file']
    file_handler_config_level = getattr(logging, file_handler_config["level"])
    if file_handler_config['enabled'] is True:
        handlers_config['file'] = {
            "class": "logging.handlers.RotatingFileHandler",
            "formatter": "file",
            "level": file_handler_config_level,
            "filename": app_config.paths["log"] / file_handler_config["filename"],
            "maxBytes": file_handler_config["maxBytes"],  # 0.5 Mb
            "backupCount": file_handler_config["backupCount"]
        }

    mindsdb_log_level = get_mindsdb_log_level()

    logging_config = dict(
        version=1,
        formatters={
            "default": {"()": ColorFormatter},
            "json": {"()": JsonFormatter},
            "file": {
                "format": "%(asctime)s %(processName)15s %(levelname)-8s %(name)s: %(message)s"
            }
        },
        handlers=handlers_config,
        loggers={
            "": {  # root logger
                "handlers": list(handlers_config.keys()),
                "level": mindsdb_log_level,
            },
            "__main__": {
                "level": mindsdb_log_level,
            },
            "mindsdb": {
                "level": mindsdb_log_level,
            },
            "alembic": {
                "level": mindsdb_log_level,
            },
        },
    )

    dictConfig(logging_config)


# I would prefer to leave code to use logging.getLogger(), but there are a lot of complicated situations
# in MindsDB with processes being spawned that require logging to be configured again in a lot of cases.
# Using a custom logger-getter like this lets us do that logic here, once.
def getLogger(name=None):
    """
    Get a new logger, configuring logging first if it hasn't been done yet.
    """
    global logging_initialized
    if not logging_initialized:
        configure_logging()
        logging_initialized = True

    return logging.getLogger(name)



from contextlib import asynccontextmanager
from collections.abc import AsyncIterator
from typing import Optional, Dict, Any
from dataclasses import dataclass

from mcp.server.fastmcp import FastMCP
from mindsdb.api.mysql.mysql_proxy.classes.fake_mysql_proxy import FakeMysqlProxy
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE as SQL_RESPONSE_TYPE
from mindsdb.utilities import log
from mindsdb.utilities.config import Config
from mindsdb.interfaces.storage import db

logger = log.getLogger(__name__)


@dataclass
class AppContext:
    db: Any


@asynccontextmanager
async def app_lifespan(server: FastMCP) -> AsyncIterator[AppContext]:
    """Manage application lifecycle with type-safe context"""
    # Initialize on startup
    db.init()
    try:
        yield AppContext(db=db)
    finally:
        # TODO: We need better way to handle this in storage/db.py
        pass


# Configure server with lifespan
mcp = FastMCP(
    "MindsDB",
    lifespan=app_lifespan,
    dependencies=["mindsdb"]  # Add any additional dependencies
)
# MCP Queries
LISTING_QUERY = "SHOW DATABASES"


@mcp.tool()
def query(query: str, context: Optional[Dict] = None) -> Dict[str, Any]:
    """
    Execute a SQL query against MindsDB

    Args:
        query: The SQL query to execute
        context: Optional context parameters for the query

    Returns:
        Dict containing the query results or error information
    """

    if context is None:
        context = {}

    logger.debug(f'Incoming MCP query: {query}')

    mysql_proxy = FakeMysqlProxy()
    mysql_proxy.set_context(context)

    try:
        result = mysql_proxy.process_query(query)

        if result.type == SQL_RESPONSE_TYPE.OK:
            return {"type": SQL_RESPONSE_TYPE.OK}

        if result.type == SQL_RESPONSE_TYPE.TABLE:
            return {
                "type": SQL_RESPONSE_TYPE.TABLE,
                "data": result.data.to_lists(json_types=True),
                "column_names": [
                    x["alias"] or x["name"] if "alias" in x else x["name"]
                    for x in result.columns
                ],
            }
        else:
            return {
                "type": SQL_RESPONSE_TYPE.ERROR,
                "error_code": 0,
                "error_message": "Unknown response type"
            }

    except Exception as e:
        logger.error(f"Error processing query: {str(e)}")
        return {
            "type": SQL_RESPONSE_TYPE.ERROR,
            "error_code": 0,
            "error_message": str(e)
        }


@mcp.tool()
def list_databases() -> Dict[str, Any]:
    """
    List all databases in MindsDB along with their tables

    Returns:
        Dict containing the list of databases and their tables
    """

    mysql_proxy = FakeMysqlProxy()

    try:
        result = mysql_proxy.process_query(LISTING_QUERY)
        if result.type == SQL_RESPONSE_TYPE.ERROR:
            return {
                "type": "error",
                "error_code": result.error_code,
                "error_message": result.error_message,
            }

        elif result.type == SQL_RESPONSE_TYPE.OK:
            return {"type": "ok"}

        elif result.type == SQL_RESPONSE_TYPE.TABLE:
            data = result.data.to_lists(json_types=True)
            return data

    except Exception as e:
        return {
            "type": "error",
            "error_code": 0,
            "error_message": str(e),
        }


def start(*args, **kwargs):
    """Start the MCP server
    Args:
        host (str): Host to bind to
        port (int): Port to listen on
    """
    config = Config()
    port = int(config['api'].get('mcp', {}).get('port', 47337))
    host = config['api'].get('mcp', {}).get('host', '127.0.0.1')

    logger.info(f"Starting MCP server on {host}:{port}")
    mcp.settings.host = host
    mcp.settings.port = port

    try:
        mcp.run(transport="sse")  # Use SSE transport instead of stdio
    except Exception as e:
        logger.error(f"Error starting MCP server: {str(e)}")
        raise


if __name__ == "__main__":
    start()



"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""

import atexit
import base64
import os
import select
import socket
import socketserver as SocketServer
import ssl
import struct
import sys
import tempfile
import traceback
from functools import partial
from typing import Dict, List, Optional
from dataclasses import dataclass

from numpy import dtype as np_dtype
from pandas.api import types as pd_types

from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import NULL_VALUE
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_datum import Datum

import mindsdb.utilities.hooks as hooks
import mindsdb.utilities.profiler as profiler
from mindsdb.api.mysql.mysql_proxy.classes.client_capabilities import ClentCapabilities
from mindsdb.api.mysql.mysql_proxy.classes.server_capabilities import (
    server_capabilities,
)
from mindsdb.api.mysql.mysql_proxy.classes.sql_statement_parser import (
    SqlStatementParser,
)
from mindsdb.api.executor.controllers import SessionController
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packet import Packet
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packets import (
    BinaryResultsetRowPacket,
    ColumnCountPacket,
    ColumnDefenitionPacket,
    CommandPacket,
    EofPacket,
    ErrPacket,
    FastAuthFail,
    HandshakePacket,
    HandshakeResponsePacket,
    OkPacket,
    PasswordAnswer,
    ResultsetRowPacket,
    STMTPrepareHeaderPacket,
    SwitchOutPacket,
    SwitchOutResponse,
)
from mindsdb.api.mysql.mysql_proxy.executor import Executor
from mindsdb.api.mysql.mysql_proxy.external_libs.mysql_scramble import (
    scramble as scramble_func,
)
from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import (
    CAPABILITIES,
    CHARSET_NUMBERS,
    COMMANDS,
    DEFAULT_AUTH_METHOD,
    ERR,
    SERVER_STATUS,
    TYPES,
    getConstName,
)
from mindsdb.api.executor.data_types.answer import ExecuteAnswer
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE
from mindsdb.api.mysql.mysql_proxy.utilities import (
    ErWrongCharset,
    SqlApiException,
)
from mindsdb.api.executor import exceptions as exec_exc

from mindsdb.api.common.check_auth import check_auth
from mindsdb.api.mysql.mysql_proxy.utilities.lightwood_dtype import dtype
from mindsdb.utilities import log
from mindsdb.utilities.config import config
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.otel.metric_handlers import get_query_request_counter
from mindsdb.utilities.wizards import make_ssl_cert

logger = log.getLogger(__name__)


def empty_fn():
    pass


@dataclass
class SQLAnswer:
    resp_type: RESPONSE_TYPE = RESPONSE_TYPE.OK
    columns: Optional[List[Dict]] = None
    data: Optional[List[Dict]] = None   # resultSet ?
    status: Optional[int] = None
    state_track: Optional[List[List]] = None
    error_code: Optional[int] = None
    error_message: Optional[str] = None
    affected_rows: Optional[int] = None

    @property
    def type(self):
        return self.resp_type


class MysqlProxy(SocketServer.BaseRequestHandler):
    """
    The Main Server controller class
    """

    @staticmethod
    def server_close(srv):
        srv.server_close()

    def __init__(self, request, client_address, server):
        self.charset = "utf8"
        self.charset_text_type = CHARSET_NUMBERS["utf8_general_ci"]
        self.session = None
        self.client_capabilities = None
        self.connection_id = None
        super().__init__(request, client_address, server)

    def init_session(self):
        logger.debug(
            "New connection [{ip}:{port}]".format(
                ip=self.client_address[0], port=self.client_address[1]
            )
        )

        if self.server.connection_id >= 65025:
            self.server.connection_id = 0
        self.server.connection_id += 1
        self.connection_id = self.server.connection_id
        self.session = SessionController(api_type='sql')

        if hasattr(self.server, "salt") and isinstance(self.server.salt, str):
            self.salt = self.server.salt
        else:
            self.salt = base64.b64encode(os.urandom(15)).decode()

        self.socket = self.request
        self.logging = logger

        self.current_transaction = None

        logger.debug("session salt: {salt}".format(salt=self.salt))

    def handshake(self):
        def switch_auth(method="mysql_native_password"):
            self.packet(SwitchOutPacket, seed=self.salt, method=method).send()
            switch_out_answer = self.packet(SwitchOutResponse)
            switch_out_answer.get()
            password = switch_out_answer.password
            if method == "mysql_native_password" and len(password) == 0:
                password = scramble_func("", self.salt)
            return password

        def get_fast_auth_password():
            logger.debug("Asking for fast auth password")
            self.packet(FastAuthFail).send()
            password_answer = self.packet(PasswordAnswer)
            password_answer.get()
            try:
                password = password_answer.password.value.decode()
            except Exception:
                logger.warning("error: no password in Fast Auth answer")
                self.packet(
                    ErrPacket,
                    err_code=ERR.ER_PASSWORD_NO_MATCH,
                    msg="Is not password in connection query.",
                ).send()
                return None
            return password

        username = None
        password = None

        logger.debug("send HandshakePacket")
        self.packet(HandshakePacket).send()

        handshake_resp = self.packet(HandshakeResponsePacket)
        handshake_resp.get()
        if handshake_resp.length == 0:
            logger.debug("HandshakeResponsePacket empty")
            self.packet(OkPacket).send()
            return False
        self.client_capabilities = ClentCapabilities(handshake_resp.capabilities.value)

        client_auth_plugin = handshake_resp.client_auth_plugin.value.decode()

        self.session.is_ssl = False

        if handshake_resp.type == "SSLRequest":
            logger.debug("switch to SSL")
            self.session.is_ssl = True

            ssl_context = ssl.SSLContext()
            ssl_context.load_cert_chain(self.server.cert_path)
            ssl_socket = ssl_context.wrap_socket(
                self.socket, server_side=True, do_handshake_on_connect=True
            )

            self.socket = ssl_socket
            handshake_resp = self.packet(HandshakeResponsePacket)
            handshake_resp.get()
            client_auth_plugin = handshake_resp.client_auth_plugin.value.decode()

        username = handshake_resp.username.value.decode()

        if client_auth_plugin != DEFAULT_AUTH_METHOD:
            if client_auth_plugin == "mysql_native_password":
                password = switch_auth("mysql_native_password")
            else:
                new_method = (
                    "caching_sha2_password"
                    if client_auth_plugin == "caching_sha2_password"
                    else "mysql_native_password"
                )

                if (
                    new_method == "caching_sha2_password"
                    and self.session.is_ssl is False
                ):
                    logger.warning(
                        f"Check auth, user={username}, ssl={self.session.is_ssl}, auth_method={client_auth_plugin}: "
                        "error: cant switch to caching_sha2_password without SSL"
                    )
                    self.packet(
                        ErrPacket,
                        err_code=ERR.ER_PASSWORD_NO_MATCH,
                        msg="caching_sha2_password without SSL not supported",
                    ).send()
                    return False

                logger.debug(
                    f"Check auth, user={username}, ssl={self.session.is_ssl}, auth_method={client_auth_plugin}: "
                    f"switch auth method to {new_method}"
                )
                password = switch_auth(new_method)

                if new_method == "caching_sha2_password":
                    if password == b"\x00":
                        password = ""
                    else:
                        password = get_fast_auth_password()
        elif "caching_sha2_password" in client_auth_plugin:
            logger.debug(
                f"Check auth, user={username}, ssl={self.session.is_ssl}, auth_method={client_auth_plugin}: "
                "check auth using caching_sha2_password"
            )
            password = handshake_resp.enc_password.value
            if password == b"\x00":
                password = ""
            else:
                # FIXME https://github.com/mindsdb/mindsdb/issues/1374
                # if self.session.is_ssl:
                #     password = get_fast_auth_password()
                # else:
                password = switch_auth()
        elif "mysql_native_password" in client_auth_plugin:
            logger.debug(
                f"Check auth, user={username}, ssl={self.session.is_ssl}, auth_method={client_auth_plugin}: "
                "check auth using mysql_native_password"
            )
            password = handshake_resp.enc_password.value
        else:
            logger.debug(
                f"Check auth, user={username}, ssl={self.session.is_ssl}, auth_method={client_auth_plugin}: "
                "unknown method, possible ERROR. Try to switch to mysql_native_password"
            )
            password = switch_auth("mysql_native_password")

        try:
            self.session.database = handshake_resp.database.value.decode()
        except Exception:
            self.session.database = None
        logger.debug(
            f"Check auth, user={username}, ssl={self.session.is_ssl}, auth_method={client_auth_plugin}: "
            f"connecting to database {self.session.database}"
        )

        auth_data = self.server.check_auth(
            username, password, scramble_func, self.salt, ctx.company_id
        )
        if auth_data["success"]:
            self.session.username = auth_data["username"]
            self.session.auth = True
            self.packet(OkPacket).send()
            return True
        else:
            self.packet(
                ErrPacket,
                err_code=ERR.ER_PASSWORD_NO_MATCH,
                msg=f"Access denied for user {username}",
            ).send()
            logger.warning(f"Access denied for user {username}")
            return False

    def send_package_group(self, packages):
        string = b"".join([x.accum() for x in packages])
        self.socket.sendall(string)

    def answer_stmt_close(self, stmt_id):
        self.session.unregister_stmt(stmt_id)

    def send_query_answer(self, answer: SQLAnswer):
        if answer.type == RESPONSE_TYPE.TABLE:
            packages = []

            if len(answer.data) > 1000:
                # for big responses leverage pandas map function to convert data to packages
                self.send_table_packets(columns=answer.columns, data=answer.data)
            else:
                packages += self.get_table_packets(columns=answer.columns, data=answer.data.to_lists())

            if answer.status is not None:
                packages.append(self.last_packet(status=answer.status))
            else:
                packages.append(self.last_packet())
            self.send_package_group(packages)
        elif answer.type == RESPONSE_TYPE.OK:
            self.packet(OkPacket, state_track=answer.state_track, affected_rows=answer.affected_rows).send()
        elif answer.type == RESPONSE_TYPE.ERROR:
            self.packet(
                ErrPacket, err_code=answer.error_code, msg=answer.error_message
            ).send()

    def _get_column_defenition_packets(self, columns, data=None, columns_len=None):
        if data is None:
            data = []
        packets = []
        for i, column in enumerate(columns):
            logger.debug(
                "%s._get_column_defenition_packets: handling column - %s of %s type",
                self.__class__.__name__,
                column,
                type(column),
            )
            table_name = column.get("table_name", "table_name")
            column_name = column.get("name", "column_name")
            column_alias = column.get("alias", column_name)
            flags = column.get("flags", 0)
            if columns_len is not None:
                length = columns_len[i]
            else:
                if len(data) == 0:
                    length = 0xFFFF
                else:
                    length = 1
                    for row in data:
                        if isinstance(row, dict):
                            length = max(len(str(row[column_alias])), length)
                        else:
                            length = max(len(str(row[i])), length)

            packets.append(
                self.packet(
                    ColumnDefenitionPacket,
                    schema=column.get("database", "mindsdb_schema"),
                    table_alias=column.get("table_alias", table_name),
                    table_name=table_name,
                    column_alias=column_alias,
                    column_name=column_name,
                    column_type=column["type"],
                    charset=column.get("charset", CHARSET_NUMBERS["utf8_unicode_ci"]),
                    max_length=length,
                    flags=flags,
                )
            )
        return packets

    def get_table_packets(self, columns, data, status=0):
        # TODO remove columns order
        packets = [self.packet(ColumnCountPacket, count=len(columns))]
        packets.extend(self._get_column_defenition_packets(columns, data))

        if self.client_capabilities.DEPRECATE_EOF is False:
            packets.append(self.packet(EofPacket, status=status))

        packets += [self.packet(ResultsetRowPacket, data=x) for x in data]
        return packets

    def send_table_packets(self, columns, data, status=0):
        # text protocol, convert all to string and serialize as packages
        df = data.get_raw_df()

        def apply_f(v):
            if v is None:
                return NULL_VALUE
            if not isinstance(v, str):
                v = str(v)
            return Datum.serialize_str(v)

        # get column max size
        # column_len is used by mysql client to determine width of columns, so it is not mandatory
        # to get exactly max value. We can approximate them by sample.
        columns_len = None
        if len(df) > 0:
            sample = df.head(100)
            columns_len = []
            for column in sample.columns:
                try:
                    columns_len.append(sample[column].astype(str).str.len().max())
                except Exception:
                    columns_len.append(1)

        # columns packages
        packets = [self.packet(ColumnCountPacket, count=len(columns))]

        packets.extend(self._get_column_defenition_packets(columns, columns_len=columns_len))

        if self.client_capabilities.DEPRECATE_EOF is False:
            packets.append(self.packet(EofPacket, status=status))
        self.send_package_group(packets)

        chunk_size = 100
        for start in range(0, len(df), chunk_size):
            string = b"".join([
                self.packet(body=body, length=len(body)).accum()
                for body in df[start:start + chunk_size].applymap(apply_f).values.sum(axis=1)
            ])
            self.socket.sendall(string)

    def decode_utf(self, text):
        try:
            return text.decode("utf-8")
        except Exception:
            raise ErWrongCharset(f"SQL contains non utf-8 values: {text}")

    def is_cloud_connection(self):
        """Determine source of connection. Must be call before handshake.
        Idea based on: real mysql connection does not send anything before server handshake, so
        soket should be in 'out' state. In opposite, clout connection sends '0000' right after
        connection. '0000' selected because in real mysql connection it should be lenght of package,
        and it can not be 0.
        """
        is_cloud = config.get("cloud", False)

        if sys.platform != "linux" or is_cloud is False:
            return {"is_cloud": False}

        read_poller = select.poll()
        read_poller.register(self.request, select.POLLIN)
        events = read_poller.poll(30)

        if len(events) == 0:
            return {"is_cloud": False}

        first_byte = self.request.recv(4, socket.MSG_PEEK)
        if first_byte == b"\x00\x00\x00\x00":
            self.request.recv(4)
            client_capabilities = self.request.recv(8)
            client_capabilities = struct.unpack("L", client_capabilities)[0]

            company_id = self.request.recv(4)
            company_id = struct.unpack("I", company_id)[0]

            user_class = self.request.recv(1)
            user_class = struct.unpack("B", user_class)[0]
            email_confirmed = 1
            if user_class > 1:
                email_confirmed = (user_class >> 2) & 1
            user_class = user_class & 3

            database_name_len = self.request.recv(2)
            database_name_len = struct.unpack("H", database_name_len)[0]

            database_name = ""
            if database_name_len > 0:
                database_name = self.request.recv(database_name_len).decode()

            return {
                "is_cloud": True,
                "client_capabilities": client_capabilities,
                "company_id": company_id,
                "user_class": user_class,
                "database": database_name,
                "email_confirmed": email_confirmed,
            }

        return {"is_cloud": False}

    def to_mysql_columns(self, columns_list):
        """Converts raw columns data into convinient format(list of lists) for the futher usage.
        Plus, it is also converts column types into internal ones."""

        result = []

        database = (
            None if self.session.database == "" else self.session.database.lower()
        )
        for column_record in columns_list:

            field_type = column_record.type

            column_type = TYPES.MYSQL_TYPE_VAR_STRING
            # is already in mysql protocol type?
            if isinstance(field_type, int):
                column_type = field_type
            # pandas checks
            elif isinstance(field_type, np_dtype):
                if pd_types.is_integer_dtype(field_type):
                    column_type = TYPES.MYSQL_TYPE_LONG
                elif pd_types.is_numeric_dtype(field_type):
                    column_type = TYPES.MYSQL_TYPE_DOUBLE
                elif pd_types.is_datetime64_any_dtype(field_type):
                    column_type = TYPES.MYSQL_TYPE_DATETIME
            # lightwood checks
            elif field_type == dtype.date:
                column_type = TYPES.MYSQL_TYPE_DATE
            elif field_type == dtype.datetime:
                column_type = TYPES.MYSQL_TYPE_DATETIME
            elif field_type == dtype.float:
                column_type = TYPES.MYSQL_TYPE_DOUBLE
            elif field_type == dtype.integer:
                column_type = TYPES.MYSQL_TYPE_LONG

            result.append(
                {
                    "database": column_record.database or database,
                    #  TODO add 'original_table'
                    "table_name": column_record.table_name,
                    "name": column_record.name,
                    "alias": column_record.alias or column_record.name,
                    # NOTE all work with text-type, but if/when wanted change types to real,
                    # it will need to check all types casts in BinaryResultsetRowPacket
                    "type": column_type,
                }
            )
        return result

    @profiler.profile()
    def process_query(self, sql):
        executor = Executor(session=self.session, sqlserver=self)
        executor.query_execute(sql)
        executor_answer = executor.executor_answer

        if executor_answer.data is None:
            resp = SQLAnswer(
                resp_type=RESPONSE_TYPE.OK,
                state_track=executor_answer.state_track,
                affected_rows=executor_answer.affected_rows
            )
        else:
            resp = SQLAnswer(
                resp_type=RESPONSE_TYPE.TABLE,
                state_track=executor_answer.state_track,
                columns=self.to_mysql_columns(executor_answer.data.columns),
                data=executor_answer.data,
                status=executor.server_status,
                affected_rows=executor_answer.affected_rows
            )

        # Increment the counter and include metadata in attributes
        metadata = ctx.metadata(query=sql)
        query_request_counter = get_query_request_counter()
        query_request_counter.add(1, metadata)

        return resp

    def answer_stmt_prepare(self, sql):
        executor = Executor(session=self.session, sqlserver=self)
        stmt_id = self.session.register_stmt(executor)

        executor.stmt_prepare(sql)

        packages = [
            self.packet(
                STMTPrepareHeaderPacket,
                stmt_id=stmt_id,
                num_columns=len(executor.columns),
                num_params=len(executor.params),
            )
        ]

        if len(executor.params) > 0:
            parameters_def = self.to_mysql_columns(executor.params)
            packages.extend(self._get_column_defenition_packets(parameters_def))
            if self.client_capabilities.DEPRECATE_EOF is False:
                status = sum([SERVER_STATUS.SERVER_STATUS_AUTOCOMMIT])
                packages.append(self.packet(EofPacket, status=status))

        if len(executor.columns) > 0:
            columns_def = self.to_mysql_columns(executor.columns)
            packages.extend(self._get_column_defenition_packets(columns_def))

            if self.client_capabilities.DEPRECATE_EOF is False:
                status = sum([SERVER_STATUS.SERVER_STATUS_AUTOCOMMIT])
                packages.append(self.packet(EofPacket, status=status))

        self.send_package_group(packages)

    def answer_stmt_execute(self, stmt_id, parameters):
        prepared_stmt = self.session.prepared_stmts[stmt_id]
        executor: Executor = prepared_stmt["statement"]

        executor.stmt_execute(parameters)

        executor_answer: ExecuteAnswer = executor.executor_answer

        if executor_answer.data is None:
            resp = SQLAnswer(
                resp_type=RESPONSE_TYPE.OK, state_track=executor_answer.state_track
            )
            return self.send_query_answer(resp)

        # TODO prepared_stmt['type'] == 'lock' is not used but it works
        columns_def = self.to_mysql_columns(executor_answer.data.columns)
        packages = [self.packet(ColumnCountPacket, count=len(columns_def))]

        packages.extend(self._get_column_defenition_packets(columns_def))

        if self.client_capabilities.DEPRECATE_EOF is False:
            packages.append(self.packet(EofPacket, status=0x0062))

        # send all
        for row in executor_answer.data.to_lists():
            packages.append(
                self.packet(BinaryResultsetRowPacket, data=row, columns=columns_def)
            )

        server_status = executor.server_status or 0x0002
        packages.append(self.last_packet(status=server_status))
        prepared_stmt["fetched"] += len(executor_answer.data)

        return self.send_package_group(packages)

    def answer_stmt_fetch(self, stmt_id, limit):
        prepared_stmt = self.session.prepared_stmts[stmt_id]
        executor = prepared_stmt["statement"]
        fetched = prepared_stmt["fetched"]
        executor_answer: ExecuteAnswer = executor.executor_answer

        if executor_answer.data is None:
            resp = SQLAnswer(
                resp_type=RESPONSE_TYPE.OK, state_track=executor_answer.state_track
            )
            return self.send_query_answer(resp)

        packages = []
        columns = self.to_mysql_columns(executor_answer.data.columns)
        for row in executor_answer.data[fetched:limit].to_lists():
            packages.append(
                self.packet(BinaryResultsetRowPacket, data=row, columns=columns)
            )

        prepared_stmt["fetched"] += len(executor_answer.data[fetched:limit])

        if len(executor_answer.data) <= limit + fetched:
            status = sum(
                [
                    SERVER_STATUS.SERVER_STATUS_AUTOCOMMIT,
                    SERVER_STATUS.SERVER_STATUS_LAST_ROW_SENT,
                ]
            )
        else:
            status = sum(
                [
                    SERVER_STATUS.SERVER_STATUS_AUTOCOMMIT,
                    SERVER_STATUS.SERVER_STATUS_CURSOR_EXISTS,
                ]
            )

        packages.append(self.last_packet(status=status))
        self.send_package_group(packages)

    def handle(self):
        """
        Handle new incoming connections
        :return:
        """
        ctx.set_default()

        self.server.hook_before_handle()

        logger.debug("handle new incoming connection")
        cloud_connection = self.is_cloud_connection()

        ctx.company_id = cloud_connection.get("company_id")

        self.init_session()
        if cloud_connection["is_cloud"] is False:
            if self.handshake() is False:
                return
        else:
            ctx.user_class = cloud_connection["user_class"]
            ctx.email_confirmed = cloud_connection["email_confirmed"]
            self.client_capabilities = ClentCapabilities(
                cloud_connection["client_capabilities"]
            )
            self.session.database = cloud_connection["database"]
            self.session.username = "cloud"
            self.session.auth = True

        while True:
            logger.debug("Got a new packet")
            p = self.packet(CommandPacket)

            try:
                success = p.get()
            except Exception:
                logger.error("Session closed, on packet read error")
                logger.error(traceback.format_exc())
                return

            if success is False:
                logger.debug("Session closed by client")
                return

            logger.debug(
                "Command TYPE: {type}".format(type=getConstName(COMMANDS, p.type.value))
            )

            command_names = {
                COMMANDS.COM_QUERY: "COM_QUERY",
                COMMANDS.COM_STMT_PREPARE: "COM_STMT_PREPARE",
                COMMANDS.COM_STMT_EXECUTE: "COM_STMT_EXECUTE",
                COMMANDS.COM_STMT_FETCH: "COM_STMT_FETCH",
                COMMANDS.COM_STMT_CLOSE: "COM_STMT_CLOSE",
                COMMANDS.COM_QUIT: "COM_QUIT",
                COMMANDS.COM_INIT_DB: "COM_INIT_DB",
                COMMANDS.COM_FIELD_LIST: "COM_FIELD_LIST",
            }

            command_name = command_names.get(p.type.value, f"UNKNOWN {p.type.value}")
            sql = None
            response = None
            error_type = None
            error_code = None
            error_text = None
            error_traceback = None

            try:
                if p.type.value == COMMANDS.COM_QUERY:
                    sql = self.decode_utf(p.sql.value)
                    sql = SqlStatementParser.clear_sql(sql)
                    logger.debug(f'Incoming query: {sql}')
                    profiler.set_meta(
                        query=sql, api="mysql", environment=config.get("environment")
                    )
                    with profiler.Context("mysql_query_processing"):
                        response = self.process_query(sql)
                elif p.type.value == COMMANDS.COM_STMT_PREPARE:
                    sql = self.decode_utf(p.sql.value)
                    self.answer_stmt_prepare(sql)
                elif p.type.value == COMMANDS.COM_STMT_EXECUTE:
                    self.answer_stmt_execute(p.stmt_id.value, p.parameters)
                elif p.type.value == COMMANDS.COM_STMT_FETCH:
                    self.answer_stmt_fetch(p.stmt_id.value, p.limit.value)
                elif p.type.value == COMMANDS.COM_STMT_CLOSE:
                    self.answer_stmt_close(p.stmt_id.value)
                elif p.type.value == COMMANDS.COM_QUIT:
                    logger.debug("Session closed, on client disconnect")
                    self.session = None
                    break
                elif p.type.value == COMMANDS.COM_INIT_DB:
                    new_database = p.database.value.decode()

                    executor = Executor(session=self.session, sqlserver=self)
                    executor.change_default_db(new_database)

                    response = SQLAnswer(RESPONSE_TYPE.OK)
                elif p.type.value == COMMANDS.COM_FIELD_LIST:
                    # this command is deprecated, but console client still use it.
                    response = SQLAnswer(RESPONSE_TYPE.OK)
                elif p.type.value == COMMANDS.COM_STMT_RESET:
                    response = SQLAnswer(RESPONSE_TYPE.OK)
                else:
                    logger.warning("Command has no specific handler, return OK msg")
                    logger.debug(str(p))
                    # p.pprintPacket() TODO: Make a version of print packet
                    # that sends it to debug instead
                    response = SQLAnswer(RESPONSE_TYPE.OK)

            except SqlApiException as e:
                # classified error
                error_type = "expected"

                response = SQLAnswer(
                    resp_type=RESPONSE_TYPE.ERROR,
                    error_code=e.err_code,
                    error_message=str(e),
                )

            except exec_exc.ExecutorException as e:
                # unclassified
                error_type = "expected"

                if isinstance(e, exec_exc.NotSupportedYet):
                    error_code = ERR.ER_NOT_SUPPORTED_YET
                elif isinstance(e, exec_exc.KeyColumnDoesNotExist):
                    error_code = ERR.ER_KEY_COLUMN_DOES_NOT_EXIST
                elif isinstance(e, exec_exc.TableNotExistError):
                    error_code = ERR.ER_TABLE_EXISTS_ERROR
                elif isinstance(e, exec_exc.WrongArgumentError):
                    error_code = ERR.ER_WRONG_ARGUMENTS
                elif isinstance(e, exec_exc.LogicError):
                    error_code = ERR.ER_WRONG_USAGE
                elif isinstance(e, (exec_exc.BadDbError, exec_exc.BadTableError)):
                    error_code = ERR.ER_BAD_DB_ERROR
                else:
                    error_code = ERR.ER_SYNTAX_ERROR

                response = SQLAnswer(
                    resp_type=RESPONSE_TYPE.ERROR,
                    error_code=error_code,
                    error_message=str(e),
                )
            except exec_exc.UnknownError as e:
                # unclassified
                error_type = "unexpected"

                response = SQLAnswer(
                    resp_type=RESPONSE_TYPE.ERROR,
                    error_code=ERR.ER_UNKNOWN_ERROR,
                    error_message=str(e),
                )

            except Exception as e:
                # any other exception
                error_type = "unexpected"
                error_traceback = traceback.format_exc()
                logger.error(
                    f"ERROR while executing query\n" f"{error_traceback}\n" f"{e}"
                )
                error_code = ERR.ER_SYNTAX_ERROR
                response = SQLAnswer(
                    resp_type=RESPONSE_TYPE.ERROR,
                    error_code=error_code,
                    error_message=str(e),
                )

            if response is not None:
                self.send_query_answer(response)
                if response.type == RESPONSE_TYPE.ERROR:
                    error_text = response.error_message
                    error_code = response.error_code
                    error_type = error_type or "expected"

            hooks.after_api_query(
                company_id=ctx.company_id,
                api="mysql",
                command=command_name,
                payload=sql,
                error_type=error_type,
                error_code=error_code,
                error_text=error_text,
                traceback=error_traceback,
            )

    def packet(self, packetClass=Packet, **kwargs):
        """
        Factory method for packets

        :param packetClass:
        :param kwargs:
        :return:
        """
        p = packetClass(socket=self.socket, session=self.session, proxy=self, **kwargs)
        self.session.inc_packet_sequence_number()
        return p

    def last_packet(self, status=0x0002):
        if self.client_capabilities.DEPRECATE_EOF is True:
            return self.packet(OkPacket, eof=True, status=status)
        else:
            return self.packet(EofPacket, status=status)

    def set_context(self, context):
        if "db" in context:
            self.session.database = context["db"]
        else:
            self.session.database = config.get('default_project')

        if "profiling" in context:
            self.session.profiling = context["profiling"]
        if "predictor_cache" in context:
            self.session.predictor_cache = context["predictor_cache"]
        if "show_secrets" in context:
            self.session.show_secrets = context["show_secrets"]

    def get_context(self):
        context = {
            "show_secrets": self.session.show_secrets
        }
        if self.session.database is not None:
            context["db"] = self.session.database
        if self.session.profiling is True:
            context["profiling"] = True
        if self.session.predictor_cache is False:
            context["predictor_cache"] = False

        return context

    @staticmethod
    def startProxy():
        """
        Create a server and wait for incoming connections until Ctrl-C
        """
        global logger

        cert_path = config["api"]["mysql"].get("certificate_path")
        if cert_path is None or cert_path == "":
            cert_path = tempfile.mkstemp(prefix="mindsdb_cert_", text=True)[1]
            make_ssl_cert(cert_path)
            atexit.register(lambda: os.remove(cert_path))
        elif not os.path.exists(cert_path):
            logger.error("Certificate defined in 'certificate_path' setting does not exist")

        # TODO make it session local
        server_capabilities.set(CAPABILITIES.CLIENT_SSL, config["api"]["mysql"]["ssl"])

        host = config["api"]["mysql"]["host"]
        port = int(config["api"]["mysql"]["port"])

        logger.info(f"Starting MindsDB Mysql proxy server on tcp://{host}:{port}")

        SocketServer.TCPServer.allow_reuse_address = True
        server = SocketServer.ThreadingTCPServer((host, port), MysqlProxy)
        server.mindsdb_config = config
        server.check_auth = partial(check_auth, config=config)
        server.cert_path = cert_path
        server.connection_id = 0
        server.hook_before_handle = empty_fn

        atexit.register(MysqlProxy.server_close, srv=server)

        # Activate the server; this will keep running until you
        # interrupt the program with Ctrl-C
        logger.info("Waiting for incoming connections...")
        server.serve_forever()



from mindsdb.utilities.otel import metrics, OTEL_SERVICE_NAME

_query_request_counter = None


def get_query_request_counter():
    """
    Get the query request counter
    """
    global _query_request_counter

    # Create the counter if it does not exist
    if _query_request_counter is None:
        meter_name = f"{OTEL_SERVICE_NAME}.query_service_meter"
        
        # Get the meter from the main metrics object
        meter = metrics.get_meter(meter_name)

        _query_request_counter = meter.create_counter(
            name="query_request_count",
            description="Counts the number of times the SQL query is called",
            unit="1",
        )

    return _query_request_counter



import json
import requests
from typing import Dict, Optional

import pandas as pd

from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.integrations.libs.llm.utils import get_completed_prompts


class OllamaHandler(BaseMLEngine):
    name = "ollama"
    DEFAULT_SERVE_URL = "http://localhost:11434"

    @staticmethod
    def create_validation(target, args=None, **kwargs):
        if 'using' not in args:
            raise Exception("Ollama engine requires a USING clause! Refer to its documentation for more details.")
        else:
            args = args['using']

        if 'model_name' not in args:
            raise Exception('`model_name` must be provided in the USING clause.')

        # check ollama service health
        connection = args.get('ollama_serve_url', OllamaHandler.DEFAULT_SERVE_URL)
        status = requests.get(connection + '/api/tags').status_code
        if status != 200:
            raise Exception(f"Ollama service is not working (status `{status}`). Please double check it is running and try again.")  # noqa

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        """ Pull LLM artifacts with Ollama API. """
        # arg setter
        args = args['using']
        args['target'] = target
        connection = args.get('ollama_serve_url', OllamaHandler.DEFAULT_SERVE_URL)

        def _model_check():
            """ Checks model has been pulled and that it works correctly. """
            responses = {}
            for endpoint in ['generate', 'embeddings']:
                try:
                    code = requests.post(
                        connection + f'/api/{endpoint}',
                        json={
                            'model': args['model_name'],
                            'prompt': 'Hello.',
                        }
                    ).status_code
                    responses[endpoint] = code
                except Exception:
                    responses[endpoint] = 500
            return responses

        # check model for all supported endpoints
        responses = _model_check()
        if 200 not in responses.values():
            # pull model (blocking operation) and serve
            # TODO: point to the engine storage folder instead of default location
            connection = args.get('ollama_serve_url', OllamaHandler.DEFAULT_SERVE_URL)
            requests.post(connection + '/api/pull', json={'name': args['model_name']})
            # try one last time
            responses = _model_check()
            if 200 not in responses.values():
                raise Exception(f"Ollama model `{args['model_name']}` is not working correctly. Please try pulling this model manually, check it works correctly and try again.")  # noqa
        
        supported_modes = {k: True if v == 200 else False for k, v in responses.items()}

        # check if a mode has been provided and if it is valid
        runnable_modes = [mode for mode, supported in supported_modes.items() if supported]
        if 'mode' in args:
            if args['mode'] not in runnable_modes:
                raise Exception(f"Mode `{args['mode']}` is not supported by the model `{args['model_name']}`.")
        
        # if a mode has not been provided, check if the model supports only one mode
        # if it does, set it as the default mode
        # if it supports multiple modes, set the default mode to 'generate'
        else:
            if len(runnable_modes) == 1:
                args['mode'] = runnable_modes[0]
            else:
                args['mode'] = 'generate'

        self.model_storage.json_set('args', args)

    def predict(self, df: pd.DataFrame, args: Optional[Dict] = None) -> pd.DataFrame:
        """
        Generate text completions with the local LLM.
            Args:
                df (pd.DataFrame): The input DataFrame containing data to predict.
                args (Optional[Dict]): Additional arguments for prediction parameters.
            Returns:
                pd.DataFrame: The DataFrame containing row-wise text completions.
        """
        # setup
        pred_args = args.get('predict_params', {})
        args = self.model_storage.json_get('args')
        model_name, target_col = args['model_name'], args['target']
        prompt_template = pred_args.get('prompt_template',
                                        args.get('prompt_template', 'Answer the following question: {{{{text}}}}'))

        # prepare prompts
        prompts, empty_prompt_ids = get_completed_prompts(prompt_template, df)
        df['__mdb_prompt'] = prompts

        # setup endpoint
        endpoint = args.get('mode', 'generate')

        # call llm
        completions = []
        for i, row in df.iterrows():
            if i not in empty_prompt_ids:
                connection = args.get('ollama_serve_url', OllamaHandler.DEFAULT_SERVE_URL)
                raw_output = requests.post(
                    connection + f'/api/{endpoint}',
                    json={
                        'model': model_name,
                        'prompt': row['__mdb_prompt'],
                    }
                )
                lines = raw_output.content.decode().split('\n')  # stream of output tokens

                values = []
                for line in lines:
                    if line != '':
                        info = json.loads(line)
                        if 'response' in info:
                            token = info['response']
                            values.append(token)
                        elif 'embedding' in info:
                            embedding = info['embedding']
                            values.append(embedding)

                if endpoint == 'embeddings':
                    completions.append(values)
                else:
                    completions.append(''.join(values))
            else:
                completions.append('')

        # consolidate output
        data = pd.DataFrame(completions)
        data.columns = [target_col]
        return data

    def describe(self, attribute: Optional[str] = None) -> pd.DataFrame:
        args = self.model_storage.json_get('args')
        model_name, target_col = args['model_name'], args['target']
        prompt_template = args.get('prompt_template', 'Answer the following question: {{{{text}}}}')

        if attribute == "features":
            return pd.DataFrame([[target_col, prompt_template]], columns=['target_column', 'mindsdb_prompt_template'])

        # get model info
        else:
            connection = args.get('ollama_serve_url', OllamaHandler.DEFAULT_SERVE_URL)
            model_info = requests.post(connection + '/api/show', json={'name': model_name}).json()
            return pd.DataFrame([[
                model_name,
                model_info.get('license', 'N/A'),
                model_info.get('modelfile', 'N/A'),
                model_info.get('parameters', 'N/A'),
                model_info.get('template', 'N/A'),
            ]],
                columns=[
                    'model_type',
                    'license',
                    'modelfile',
                    'parameters',
                    'ollama_base_template',
                ])



import json
import requests
from typing import Union
from google.oauth2 import service_account

from mindsdb.utilities import log

from ..exceptions import NoCredentialsException, AuthException


logger = log.getLogger(__name__)


class GoogleServiceAccountOAuth2Manager:
    def __init__(self, credentials_url: str = None, credentials_file: str = None, credentials_json: Union[dict, str] = None) -> None:
        # if no credentials provided, raise an exception
        if not any([credentials_url, credentials_file, credentials_json]):
            raise NoCredentialsException('No valid Google Service Account credentials provided.')
        self.credentials_url = credentials_url
        self.credentials_file = credentials_file
        if credentials_json:
            self.credentials_json = self._parse_credentials_json(credentials_json)
        else:
            self.credentials_json = None

    def get_oauth2_credentials(self):
        try:
            if self.credentials_url:
                creds = service_account.Credentials.from_service_account_info(self._download_credentials_file())
                return creds

            if self.credentials_file:
                creds = service_account.Credentials.from_service_account_file(self.credentials_file)
                return creds

            if self.credentials_json:
                creds = service_account.Credentials.from_service_account_info(self.credentials_json)
                return creds
        except Exception as e:
            raise AuthException(f"Authentication failed: {e}")

    def _download_credentials_file(self):
        response = requests.get(self.credentials_url)
        # raise a HTTPError if the status is 4xx or 5xx
        response.raise_for_status()

        return self._parse_credentials_json(response.json())

    def _parse_credentials_json(self, credentials_json: str) -> dict:
        if isinstance(credentials_json, str):
            try:
                # attempt to convert to JSON
                return json.loads(credentials_json)
            except json.JSONDecodeError:
                raise ValueError("Failed to parse credentials provided. Please provide a valid service account key.")
        else:
            # unescape new lines in private_key
            credentials_json['private_key'] = credentials_json['private_key'].replace('\\n', '\n')
            return credentials_json



import logging

from opentelemetry._logs import set_logger_provider
from opentelemetry.sdk._logs._internal.export import LogExporter
from opentelemetry.sdk._logs import LoggerProvider, LoggingHandler
from opentelemetry.sdk._logs.export import BatchLogRecordProcessor
from opentelemetry.sdk.resources import Resource

from mindsdb.utilities.log import get_mindsdb_log_level


def setup_logger(resource: Resource, exporter: LogExporter) -> None:
    """
    Setup OpenTelemetry logging
    """
    mindsdb_log_level = get_mindsdb_log_level()

    logger_provider = LoggerProvider(resource=resource)
    set_logger_provider(logger_provider)

    logger_provider.add_log_record_processor(BatchLogRecordProcessor(exporter))
    handler = LoggingHandler(level=mindsdb_log_level, logger_provider=logger_provider)

    # Attach OTLP handler to root logger
    logging.getLogger().addHandler(handler)



import os
from typing import Dict, Optional

from PIL import Image
import requests
import numpy as np
from io import BytesIO
import json
import textwrap
import google.generativeai as genai
import pandas as pd
from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.utilities import log
from mindsdb.utilities.config import Config
from mindsdb.integrations.libs.llm.utils import get_completed_prompts
import concurrent.futures

logger = log.getLogger(__name__)


class GoogleGeminiHandler(BaseMLEngine):
    """
    Integration with the Google generative AI Python Library
    """

    name = "google_gemini"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.default_model = "gemini-pro"
        self.default_embedding_model = "models/embedding-001"
        self.generative = True
        self.mode = "default"

    # Similiar to openai handler
    @staticmethod
    def create_validation(target, args=None, **kwargs):
        if "using" not in args:
            raise Exception(
                "Gemini engine requires a USING clause! Refer to its documentation for more details."
            )
        else:
            args = args["using"]

        if (
            len(
                set(args.keys())
                & {
                    "img_url",
                    "input_text",
                    "question_column",
                    "prompt_template",
                    "json_struct",
                    "prompt",
                }
            )
            == 0
        ):
            raise Exception(
                "One of `question_column`, `prompt_template` or `json_struct` is required for this engine."
            )

        keys_collection = [
            ["prompt_template"],
            ["question_column", "context_column"],
            ["prompt", "user_column", "assistant_column"],
            ["json_struct", "input_text"],
            ["img_url", "ctx_column"],
        ]
        for keys in keys_collection:
            if keys[0] in args and any(
                x[0] in args for x in keys_collection if x != keys
            ):
                raise Exception(
                    textwrap.dedent(
                        """\
                    Please provide one of
                        1) a `prompt_template`
                        2) a `question_column` and an optional `context_column`
                        3) a `json_struct`
                        4) a `prompt' and 'user_column' and 'assistant_column`
                        5) a `img_url` and optional `ctx_column` for mode=`vision`
                """
                    )
                )

        # for all args that are not expected, raise an error
        known_args = set()
        # flatten of keys_collection
        for keys in keys_collection:
            known_args = known_args.union(set(keys))

        # TODO: need a systematic way to maintain a list of known args
        known_args = known_args.union(
            {
                "target",
                "model_name",
                "mode",
                "title_column",
                "predict_params",
                "type",
                "max_tokens",
                "temperature",
                "api_key",
            }
        )

        unknown_args = set(args.keys()) - known_args
        if unknown_args:
            # return a list of unknown args as a string
            raise Exception(
                f"Unknown arguments: {', '.join(unknown_args)}.\n Known arguments are: {', '.join(known_args)}"
            )

    def create(self, target, args=None, **kwargs):
        args = args["using"]
        args["target"] = target
        self.model_storage.json_set("args", args)

    def predict(
        self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None
    ) -> pd.DataFrame:
        pred_args = args["predict_params"] if args else {}
        args = self.model_storage.json_get("args")
        df = df.reset_index(drop=True)

        # same as opeani handler for getting prompt template and mode
        if pred_args.get("prompt_template", False):
            base_template = pred_args[
                "prompt_template"
            ]  # override with predict-time template if available
        elif args.get("prompt_template", False):
            base_template = args["prompt_template"]
        else:
            base_template = None

        # Embedding Mode
        if args.get("mode") == "embedding":
            args["type"] = pred_args.get("type", "query")
            return self.embedding_worker(args, df)

        elif args.get("mode") == "vision":
            return self.vision_worker(args, df)

        elif args.get("mode") == "conversational":
            # Enable chat mode using
            # https://ai.google.dev/tutorials/python_quickstart#chat_conversations
            # OR
            # https://github.com/google/generative-ai-python?tab=readme-ov-file#developers-who-use-the-palm-api
            pass

        else:
            if args.get("prompt_template", False):
                prompts, empty_prompt_ids = get_completed_prompts(base_template, df)

            # Disclaimer: The following code has been adapted from the OpenAI handler.
            elif args.get("context_column", False):
                empty_prompt_ids = np.where(
                    df[[args["context_column"], args["question_column"]]]
                    .isna()
                    .all(axis=1)
                    .values
                )[0]
                contexts = list(df[args["context_column"]].apply(lambda x: str(x)))
                questions = list(df[args["question_column"]].apply(lambda x: str(x)))
                prompts = [
                    f"Give only answer for: \nContext: {c}\nQuestion: {q}\nAnswer: "
                    for c, q in zip(contexts, questions)
                ]

                # Disclaimer: The following code has been adapted from the OpenAI handler.
            elif args.get("json_struct", False):
                empty_prompt_ids = np.where(
                    df[[args["input_text"]]].isna().all(axis=1).values
                )[0]
                prompts = []
                for i in df.index:
                    if "json_struct" in df.columns:
                        if isinstance(df["json_struct"][i], str):
                            df["json_struct"][i] = json.loads(df["json_struct"][i])
                        json_struct = ""
                        for ind, val in enumerate(df["json_struct"][i].values()):
                            json_struct = json_struct + f"{ind}. {val}\n"
                    else:
                        json_struct = ""
                        for ind, val in enumerate(args["json_struct"].values()):
                            json_struct = json_struct + f"{ind + 1}. {val}\n"

                    p = textwrap.dedent(
                        f"""\
                        Using text starting after 'The text is:', give exactly {len(args['json_struct'])} answers to the questions:
                        {{{{json_struct}}}}

                        Answers should be in the same order as the questions.
                        Answer should be in form of one JSON Object eg. {"{'key':'value',..}"} where key=question and value=answer.
                        If there is no answer to the question in the text, put a -.
                        Answers should be as short as possible, ideally 1-2 words (unless otherwise specified).

                        The text is:
                        {{{{{args['input_text']}}}}}
                    """
                    )
                    p = p.replace("{{json_struct}}", json_struct)
                    for column in df.columns:
                        if column == "json_struct":
                            continue
                        p = p.replace(f"{{{{{column}}}}}", str(df[column][i]))
                    prompts.append(p)
            elif "prompt" in args:
                empty_prompt_ids = []
                prompts = list(df[args["user_column"]])
            else:
                empty_prompt_ids = np.where(
                    df[[args["question_column"]]].isna().all(axis=1).values
                )[0]
                prompts = list(df[args["question_column"]].apply(lambda x: str(x)))

        # remove prompts without signal from completion queue
        prompts = [j for i, j in enumerate(prompts) if i not in empty_prompt_ids]

        api_key = self._get_google_gemini_api_key(args)
        genai.configure(api_key=api_key)

        # called gemini model withinputs
        model = genai.GenerativeModel(args.get("model_name", self.default_model))
        results = []
        for m in prompts:
            results.append(model.generate_content(m).text)

        pred_df = pd.DataFrame(results, columns=[args["target"]])
        return pred_df

    def _get_google_gemini_api_key(self, args, strict=True):
        """
        API_KEY preference order:
            1. provided at model creation
            2. provided at engine creation
            3. GOOGLE_GENAI_API_KEY env variable
            4. google_gemini.api_key setting in config.json
        """

        if "api_key" in args:
            return args["api_key"]
        # 2
        connection_args = self.engine_storage.get_connection_args()
        if "api_key" in connection_args:
            return connection_args["api_key"]
        # 3
        api_key = os.getenv("GOOGLE_GENAI_API_KEY")
        if api_key is not None:
            return api_key
        # 4
        config = Config()
        google_gemini_config = config.get("google_gemini", {})
        if "api_key" in google_gemini_config:
            return google_gemini_config["api_key"]

        if strict:
            raise Exception(
                'Missing API key "api_key". Either re-create this ML_ENGINE specifying the `api_key` parameter,\
                 or re-create this model and pass the API key with `USING` syntax.'
            )

    def embedding_worker(self, args: Dict, df: pd.DataFrame):
        if args.get("question_column"):
            prompts = list(df[args["question_column"]].apply(lambda x: str(x)))
            if args.get("title_column", None):
                titles = list(df[args["title_column"]].apply(lambda x: str(x)))
            else:
                titles = None

            api_key = self._get_google_gemini_api_key(args)
            genai.configure(api_key=api_key)
            model_name = args.get("model_name", self.default_embedding_model)
            task_type = args.get("type")
            task_type = f"retrieval_{task_type}"

            if task_type == "retrieval_query":
                results = [
                    str(
                        genai.embed_content(
                            model=model_name, content=query, task_type=task_type
                        )["embedding"]
                    )
                    for query in prompts
                ]
            elif titles:
                results = [
                    str(
                        genai.embed_content(
                            model=model_name,
                            content=doc,
                            task_type=task_type,
                            title=title,
                        )["embedding"]
                    )
                    for title, doc in zip(titles, prompts)
                ]
            else:
                results = [
                    str(
                        genai.embed_content(
                            model=model_name, content=doc, task_type=task_type
                        )["embedding"]
                    )
                    for doc in prompts
                ]

            pred_df = pd.DataFrame(results, columns=[args["target"]])
            return pred_df
        else:
            raise Exception("Embedding mode needs a question_column")

    def vision_worker(self, args: Dict, df: pd.DataFrame):
        def get_img(url):
            # URL Validation
            response = requests.get(url)
            if response.status_code == 200 and response.headers.get(
                "content-type", ""
            ).startswith("image/"):
                return Image.open(BytesIO(response.content))
            else:
                raise Exception(f"{url} is not vaild image URL..")

        if args.get("img_url"):
            urls = list(df[args["img_url"]].apply(lambda x: str(x)))

        else:
            raise Exception("Vision mode needs a img_url")

        prompts = None
        if args.get("ctx_column"):
            prompts = list(df[args["ctx_column"]].apply(lambda x: str(x)))

        api_key = self._get_google_gemini_api_key(args)
        genai.configure(api_key=api_key)
        model = genai.GenerativeModel("gemini-pro-vision")
        with concurrent.futures.ThreadPoolExecutor() as executor:
            # Download images concurrently using ThreadPoolExecutor
            imgs = list(executor.map(get_img, urls))
        # imgs = [Image.open(BytesIO(requests.get(url).content)) for url in urls]
        if prompts:
            results = [
                model.generate_content([img, text]).text
                for img, text in zip(imgs, prompts)
            ]
        else:
            results = [model.generate_content(img).text for img in imgs]

        pred_df = pd.DataFrame(results, columns=[args["target"]])

        return pred_df

    # Disclaimer: The following code has been adapted from the OpenAI handler.
    def describe(self, attribute: Optional[str] = None) -> pd.DataFrame:

        args = self.model_storage.json_get("args")

        if attribute == "args":
            return pd.DataFrame(args.items(), columns=["key", "value"])
        elif attribute == "metadata":
            api_key = self._get_google_gemini_api_key(args)
            genai.configure(api_key=api_key)
            model_name = args.get("model_name", self.default_model)

            meta = genai.get_model(f"models/{model_name}").__dict__
            return pd.DataFrame(meta.items(), columns=["key", "value"])
        else:
            tables = ["args", "metadata"]
            return pd.DataFrame(tables, columns=["tables"])



from typing import Any, Dict, List, Optional

from pydantic import BaseModel, ConfigDict, Field


class BaseLLMConfig(BaseModel):
    # Remove 'model_' prefix from protected namespaces since Langchain constructor
    # kwargs share the same prefix.
    model_config = ConfigDict(protected_namespaces=())


# See https://api.python.langchain.com/en/latest/chat_models/langchain_community.chat_models.openai.ChatOpenAI.html#langchain_community.chat_models.openai.ChatOpenAI
# This config does not have to be exclusively used with Langchain.
class OpenAIConfig(BaseLLMConfig):
    model_name: str
    temperature: Optional[float]
    max_retries: Optional[int]
    max_tokens: Optional[int]
    openai_api_base: Optional[str]
    # Inferred from OPENAI_API_KEY if not provided.
    openai_api_key: Optional[str]
    openai_organization: Optional[str]
    request_timeout: Optional[float]


# See https://api.python.langchain.com/en/latest/chat_models/langchain_community.chat_models.anthropic.ChatAnthropic.html
# This config does not have to be exclusively used with Langchain.
class AnthropicConfig(BaseLLMConfig):
    model: str
    temperature: Optional[float]
    max_tokens: Optional[int]
    top_p: Optional[float]
    top_k: Optional[int]
    default_request_timeout: Optional[float]
    # Inferred from ANTHROPIC_API_KEY if not provided.
    anthropic_api_key: Optional[str]
    anthropic_api_url: Optional[str]


# See https://api.python.langchain.com/en/latest/chat_models/langchain_community.chat_models.anyscale.ChatAnyscale.html
# This config does not have to be exclusively used with Langchain.
class AnyscaleConfig(BaseLLMConfig):
    model_name: str
    temperature: Optional[float]
    max_retries: Optional[int]
    max_tokens: Optional[int]
    anyscale_api_base: Optional[str]
    # Inferred from ANYSCALE_API_KEY if not provided.
    anyscale_api_key: Optional[str]
    anyscale_proxy: Optional[str]
    request_timeout: Optional[float]


# See https://api.python.langchain.com/en/latest/chat_models/langchain_community.chat_models.litellm.ChatLiteLLM.html
# This config does not have to be exclusively used with Langchain.
class LiteLLMConfig(BaseLLMConfig):
    model: str
    api_base: Optional[str]
    max_retries: Optional[int]
    max_tokens: Optional[int]
    top_p: Optional[float]
    top_k: Optional[int]
    temperature: Optional[float]
    custom_llm_provider: Optional[str]
    model_kwargs: Optional[Dict[str, Any]]


# See https://api.python.langchain.com/en/latest/chat_models/langchain_community.chat_models.ollama.ChatOllama.html
# This config does not have to be exclusively used with Langchain.
class OllamaConfig(BaseLLMConfig):
    base_url: str
    model: str
    temperature: Optional[float]
    top_p: Optional[float]
    top_k: Optional[int]
    timeout: Optional[int]
    format: Optional[str]
    headers: Optional[Dict]
    num_predict: Optional[int]
    num_ctx: Optional[int]
    num_gpu: Optional[int]
    repeat_penalty: Optional[float]
    stop: Optional[List[str]]
    template: Optional[str]


class NvidiaNIMConfig(BaseLLMConfig):
    base_url: str
    model: str
    temperature: Optional[float]
    top_p: Optional[float]
    timeout: Optional[int]
    format: Optional[str]
    headers: Optional[Dict]
    num_predict: Optional[int]
    num_ctx: Optional[int]
    num_gpu: Optional[int]
    repeat_penalty: Optional[float]
    stop: Optional[List[str]]
    template: Optional[str]
    nvidia_api_key: Optional[str]


class MindsdbConfig(BaseLLMConfig):
    model_name: str
    project_name: str


# See https://python.langchain.com/api_reference/google_genai/chat_models/langchain_google_genai.chat_models.ChatGoogleGenerativeAI.html
class GoogleConfig(BaseLLMConfig):
    model: str = Field(description="Gemini model name to use (e.g., 'gemini-1.5-pro')")
    temperature: Optional[float] = Field(default=None, description="Controls randomness in responses")
    top_p: Optional[float] = Field(default=None, description="Nucleus sampling parameter")
    top_k: Optional[int] = Field(default=None, description="Number of highest probability tokens to consider")
    max_output_tokens: Optional[int] = Field(default=None, description="Maximum number of tokens to generate")
    google_api_key: Optional[str] = Field(default=None, description="API key for Google Generative AI")



from merlion.models.anomaly.forecast_based.prophet import ProphetDetectorConfig, ProphetDetector
from merlion.models.anomaly.isolation_forest import IsolationForestConfig, IsolationForest
from merlion.models.anomaly.windstats import WindStatsConfig, WindStats
from merlion.models.automl.autoprophet import AutoProphetConfig, AutoProphet
from merlion.models.automl.autosarima import AutoSarimaConfig, AutoSarima
from merlion.models.forecast.smoother import MSESConfig, MSES
from merlion.post_process.threshold import AggregateAlarms
from merlion.transform.moving_average import DifferenceTransform
from scipy.stats import norm

from merlion.models.defaults import DefaultDetectorConfig, DefaultDetector, DefaultForecaster, DefaultForecasterConfig
from merlion.utils import TimeSeries

from enum import Enum
import pandas as pd


class MerlionArguments(Enum):
    target_seq_index = "target_seq_index"
    max_forecast_steps = "max_forecast_steps"
    max_backstep = "max_backstep"
    wind_sz = "wind_sz"
    alm_threshold = "alm_threshold"
    maxiter = "maxiter"


class BaseMerlionForecastAdapter:
    TARGET_SEQ_INDEX = 0
    DEFAULT_MAX_FORECAST_STEPS = 100
    DEFAULT_MAX_BACKSTEP = 60
    DEFAULT_MAXITER = 5

    def __init__(self, **kwargs):
        self.max_forecast_steps = kwargs.get(MerlionArguments.max_forecast_steps.value, self.DEFAULT_MAX_FORECAST_STEPS)
        self.max_backstep = kwargs.get(MerlionArguments.max_backstep.value, self.DEFAULT_MAX_BACKSTEP)
        self.maxiter = kwargs.get(MerlionArguments.maxiter.value, self.DEFAULT_MAXITER)
        self.model = None

    def to_bytes(self) -> bytes:
        return self.model.to_bytes()

    def initialize_model(self, bytes):
        self.model = self.model.from_bytes(bytes)

    def to_train_dataframe(self, df: pd.DataFrame, target: str) -> pd.DataFrame:
        columns = list(df.columns.values)
        columns.remove(target)
        columns.insert(self.TARGET_SEQ_INDEX, target)
        return df[columns]

    def train(self, df: pd.DataFrame, target: str):
        df = self.to_train_dataframe(df=df, target=target)
        train_data = TimeSeries.from_pd(df)
        self.model.train(train_data=train_data)

    def predict(self, df: pd.DataFrame, target: str) -> pd.DataFrame:
        forecast_step = self.max_forecast_steps
        df = df[df.index <= self.model.last_train_time + self.model.timedelta * forecast_step]
        if len(list(df.columns.values)) == 0:
            df.loc[:, target] = 0
        predict_data = TimeSeries.from_pd(df)
        predict_pred, predict_err = self.model.forecast(time_stamps=predict_data.time_stamps)
        return self.__prepare_forecast_return(target=target, pred_ts=predict_pred, err_ts=predict_err)

    def __prepare_forecast_return(self, target: str, pred_ts: TimeSeries, err_ts: TimeSeries) -> pd.DataFrame:
        pred_df: pd.DataFrame = pred_ts.to_pd()
        if err_ts is None:
            err_df = None
        else:
            err_df = err_ts.to_pd()

        if err_df is None or target in list(err_df.columns.values):  # error and predict sometimes are same
            std = pred_df[target].std()
            pred_df[f"{target}__upper"] = pred_df[f"{target}"] + std * norm.ppf(0.975)
            pred_df[f"{target}__lower"] = pred_df[f"{target}"] + std * norm.ppf(0.025)
        else:
            pred_df = pred_df.join(err_df, how="left")
            pred_df[f"{target}__upper"] = pred_df[f"{target}"] + norm.ppf(0.975) * err_df[f"{target}_err"]
            pred_df[f"{target}__lower"] = pred_df[f"{target}"] + norm.ppf(0.025) * err_df[f"{target}_err"]
            pred_df.drop(columns=[f"{target}_err"], inplace=True)
        return pred_df


class DefaultForecasterAdapter(BaseMerlionForecastAdapter):
    # DefaultForecaster
    # reference: https://opensource.salesforce.com/Merlion/latest/merlion.models.html#
    #               merlion.models.defaults.DefaultForecaster
    def __init__(self, **kwargs):
        super(DefaultForecasterAdapter, self).__init__(**kwargs)
        self.model = DefaultForecaster(DefaultForecasterConfig(max_forecast_steps=self.max_forecast_steps,
                                                               target_seq_index=self.TARGET_SEQ_INDEX))


class SarimaForecasterAdapter(BaseMerlionForecastAdapter):
    # AutoSarima
    # reference: https://opensource.salesforce.com/Merlion/latest/merlion.models.automl.html#
    #               module-merlion.models.automl.autosarima
    def __init__(self, **kwargs):
        super(SarimaForecasterAdapter, self).__init__(**kwargs)
        config = AutoSarimaConfig(auto_pqPQ=True, auto_d=True, auto_D=True, auto_seasonality=True,
                                  approximation=True, maxiter=5)
        self.model = AutoSarima(config)


class ProphetForecasterAdapter(BaseMerlionForecastAdapter):
    # AutoProphet
    # reference: https://opensource.salesforce.com/Merlion/latest/merlion.models.automl.html#
    #               module-merlion.models.automl.autoprophet
    def __init__(self, **kwargs):
        super(ProphetForecasterAdapter, self).__init__(**kwargs)
        self.model = AutoProphet(AutoProphetConfig(max_forecast_steps=self.max_forecast_steps))


class MSESForecasterAdapter(BaseMerlionForecastAdapter):
    # MSES
    # reference: https://opensource.salesforce.com/Merlion/latest/merlion.models.forecast.html#
    #               merlion.models.forecast.smoother.MSES
    def __init__(self, **kwargs):
        super(MSESForecasterAdapter, self).__init__(**kwargs)
        self.model = MSES(MSESConfig(max_forecast_steps=self.max_forecast_steps,
                                     target_seq_index=self.TARGET_SEQ_INDEX,
                                     max_backstep=self.max_backstep))


class BaseMerlineDetectorAdapter(BaseMerlionForecastAdapter):
    DEFAULT_WIND_SZ = 60
    DEFAULT_ALM_THRESHOLD = 4

    def __init__(self, **kwargs):
        super(BaseMerlineDetectorAdapter, self).__init__(**kwargs)
        self.wind_sz = kwargs.get(MerlionArguments.wind_sz.value, self.DEFAULT_WIND_SZ)
        self.alm_threshold = kwargs.get(MerlionArguments.alm_threshold.value, self.DEFAULT_ALM_THRESHOLD)

    def train(self, df: pd.DataFrame, target: str):
        df = self.to_train_dataframe(df=df, target=target)
        train_data = TimeSeries.from_pd(df[target])
        self.model.train(train_data=train_data)

    def predict(self, df: pd.DataFrame, target: str) -> pd.DataFrame:
        predict_data = TimeSeries.from_pd(df[target])
        predict_label = self.model.get_anomaly_label(time_series=predict_data)
        pred_df = predict_label.to_pd()
        pred_df[f"{target}__anomaly_score"] = pred_df["anom_score"]
        return pred_df


class DefaultDetectorAdapter(BaseMerlineDetectorAdapter):
    # DefaultDetector
    # reference: https://opensource.salesforce.com/Merlion/latest/merlion.models.html#
    #               merlion.models.defaults.DefaultDetector
    def __init__(self, **kwargs):
        super(DefaultDetectorAdapter, self).__init__(**kwargs)
        self.model = DefaultDetector(DefaultDetectorConfig())


class IsolationForestDetectorAdapter(BaseMerlineDetectorAdapter):
    # IsolationForest
    # reference: https://opensource.salesforce.com/Merlion/latest/merlion.models.anomaly.html#
    #               merlion.models.anomaly.isolation_forest.IsolationForest
    def __init__(self, **kwargs):
        super(IsolationForestDetectorAdapter, self).__init__(**kwargs)
        self.model = IsolationForest(IsolationForestConfig())


class WindStatsDetectorAdapter(BaseMerlineDetectorAdapter):
    # WindStats
    # reference: https://opensource.salesforce.com/Merlion/latest/merlion.models.anomaly.html#
    #               merlion.models.anomaly.windstats.WindStats
    def __init__(self, **kwargs):
        super(WindStatsDetectorAdapter, self).__init__(**kwargs)
        config = WindStatsConfig(wind_sz=self.wind_sz, threshold=AggregateAlarms(alm_threshold=self.alm_threshold))
        self.model = WindStats(config)


class ProphetDetectorAdapter(BaseMerlineDetectorAdapter):
    # ProphetDetector
    # reference: https://opensource.salesforce.com/Merlion/latest/merlion.models.anomaly.forecast_based.html#
    #               merlion.models.anomaly.forecast_based.prophet.ProphetDetector
    def __init__(self, **kwargs):
        super(ProphetDetectorAdapter, self).__init__(**kwargs)
        config = ProphetDetectorConfig(transform=DifferenceTransform())
        self.model = ProphetDetector(config)



import pandas as pd
from typing import List
from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.utilities.handlers.query_utilities import (
    SELECTQueryParser,
    SELECTQueryExecutor,
)
from mindsdb.utilities import log
from mindsdb_sql_parser import ast
import zenpy

logger = log.getLogger(__name__)


class ZendeskUsersTable(APITable):
    """Zendesk Users Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the zendesk list users API

        Parameters
        ----------
        query : ast.Select
            Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Zendesk users

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'users',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = (
            select_statement_parser.parse_query()
        )

        subset_where_conditions = {}
        for op, arg1, arg2 in where_conditions:
            if arg1 in self.get_columns():
                if op != '=':
                    raise NotImplementedError(f"Unknown op: {op}. Only '=' is supported.")
                subset_where_conditions[arg1] = arg2

        count = 0
        result = self.handler.zen_client.users(**subset_where_conditions)
        response = []
        if isinstance(result, zenpy.lib.generator.BaseResultGenerator):
            while count <= result_limit:
                response.append(result.next().to_dict())
                count += 1
        else:
            response.append(result.to_dict())

        df = pd.DataFrame(response, columns=self.get_columns())

        select_statement_executor = SELECTQueryExecutor(
            df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions,
            result_limit
        )

        df = select_statement_executor.execute_query()

        return df

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "active", "alias", "chat_only", "created_at", "custom_role_id",
            "details", "email", "external_id", "id", "last_login_at",
            "locale", "locale_id", "moderator", "name", "notes",
            "only_private_comments", "organization_id", "phone", "photo",
            "restricted_agent", "role", "shared", "shared_agent",
            "signature", "suspended", "tags", "ticket_restriction",
            "time_zone", "two_factor_auth_enabled", "updated_at", "url",
            "verified", "iana_time_zone", "shared_phone_number", "role_type",
            "default_group_id", "report_csv"
        ]


class ZendeskTicketsTable(APITable):
    """Zendesk tickets Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the zendesk tickets API

        Parameters
        ----------
        query : ast.Select
            Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Ticket ID Data

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'tickets',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = (
            select_statement_parser.parse_query()
        )

        subset_where_conditions = {}
        for op, arg1, arg2 in where_conditions:
            if arg1 in self.get_columns():
                if op != '=':
                    raise NotImplementedError(f"Unknown op: {op}. Only '=' is supported.")
                subset_where_conditions[arg1] = arg2

        count = 0
        result = self.handler.zen_client.tickets(**subset_where_conditions)
        response = []
        if isinstance(result, zenpy.lib.generator.BaseResultGenerator):
            while count <= result_limit:
                response.append(result.next().to_dict())
                count += 1
        else:
            response.append(result.to_dict())

        df = pd.DataFrame(response, columns=self.get_columns())

        select_statement_executor = SELECTQueryExecutor(
            df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions,
            result_limit
        )

        df = select_statement_executor.execute_query()

        return df

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "assignee_id", "brand_id", "collaborator_ids", "created_at",
            "custom_fields", "description", "due_at", "external_id",
            "fields", "forum_topic_id", "group_id", "has_incidents", "id",
            "organization_id", "priority", "problem_id", "raw_subject",
            "recipient", "requester_id", "sharing_agreement_ids", "status",
            "subject", "submitter_id", "tags", "type", "updated_at", "url",
            "generated_timestamp", "follower_ids", "email_cc_ids", "is_public",
            "custom_status_id", "followup_ids", "ticket_form_id",
            "allow_channelback", "allow_attachments", "from_messaging_channel",
            "satisfaction_rating.assignee_id", "satisfaction_rating.created_at",
            "satisfaction_rating.group_id", "satisfaction_rating.id",
            "satisfaction_rating.requester_id", "satisfaction_rating.score",
            "satisfaction_rating.ticket_id", "satisfaction_rating.updated_at",
            "satisfaction_rating.url", "via.channel", "via.source.rel"
        ]


class ZendeskTriggersTable(APITable):
    """Zendesk Triggers Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the zendesk triggers API

        Parameters
        ----------
        query : ast.Select
            Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Trigger Data

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'triggers',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = (
            select_statement_parser.parse_query()
        )

        subset_where_conditions = {}
        for op, arg1, arg2 in where_conditions:
            if arg1 in self.get_columns():
                if op != '=':
                    raise NotImplementedError(f"Unknown op: {op}. Only '=' is supported.")
                subset_where_conditions[arg1] = arg2

        count = 0
        result = self.handler.zen_client.triggers(**subset_where_conditions)
        response = []
        if isinstance(result, zenpy.lib.generator.BaseResultGenerator):
            while count <= result_limit:
                response.append(result.next().to_dict())
                count += 1
        else:
            response.append(result.to_dict())

        df = pd.DataFrame(response, columns=self.get_columns())

        select_statement_executor = SELECTQueryExecutor(
            df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions,
            result_limit
        )

        df = select_statement_executor.execute_query()

        return df

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "actions", "active", "description", "id", "position", "title",
            "url", "updated_at", "created_at", "default", "raw_title",
            "category_id", "conditions.all", "conditions.any"
        ]


class ZendeskActivitiesTable(APITable):
    """Zendesk Activities Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the zendesk activities API

        Parameters
        ----------
        query : ast.Select
            Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Activity list Data

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'activities',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = (
            select_statement_parser.parse_query()
        )

        subset_where_conditions = {}
        for op, arg1, arg2 in where_conditions:
            if arg1 in self.get_columns():
                if op != '=':
                    raise NotImplementedError(f"Unknown op: {op}. Only '=' is supported.")
                subset_where_conditions[arg1] = arg2

        count = 0
        result = self.handler.zen_client.activities(**subset_where_conditions)
        response = []
        if isinstance(result, zenpy.lib.generator.BaseResultGenerator):
            while count <= result_limit:
                response.append(result.next().to_dict())
                count += 1
        else:
            response.append(result.to_dict())

        df = pd.DataFrame(response, columns=self.get_columns())

        select_statement_executor = SELECTQueryExecutor(
            df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions,
            result_limit
        )

        df = select_statement_executor.execute_query()

        return df

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "created_at",
            "id",
            "title",
            "updated_at",
            "url",
            "verb",
            "user_id",
            "actor_id",
            "actor.id",
            "actor.url",
            "actor.name",
            "actor.email",
            "actor.created_at",
            "actor.updated_at",
            "actor.time_zone",
            "actor.iana_time_zone",
            "actor.phone",
            "actor.shared_phone_number",
            "actor.photo",
            "actor.locale_id",
            "actor.locale",
            "actor.organization_id",
            "actor.role",
            "actor.verified",
            "actor.external_id",
            "actor.tags",
            "actor.alias",
            "actor.active",
            "actor.shared",
            "actor.shared_agent",
            "actor.last_login_at",
            "actor.two_factor_auth_enabled",
            "actor.signature",
            "actor.details",
            "actor.notes",
            "actor.role_type",
            "actor.custom_role_id",
            "actor.moderator",
            "actor.ticket_restriction",
            "actor.only_private_comments",
            "actor.restricted_agent",
            "actor.suspended",
            "actor.default_group_id",
            "actor.report_csv",
            "user.active",
            "user.alias",
            "user.chat_only",
            "user.created_at",
            "user.custom_role_id",
            "user.details",
            "user.email",
            "user.external_id",
            "user.id",
            "user.last_login_at",
            "user.locale",
            "user.locale_id",
            "user.moderator",
            "user.name",
            "user.notes",
            "user.only_private_comments",
            "user.organization_id",
            "user.phone",
            "user.photo",
            "user.restricted_agent",
            "user.role",
            "user.shared",
            "user.shared_agent",
            "user.signature",
            "user.suspended",
            "user.tags",
            "user.ticket_restriction",
            "user.time_zone",
            "user.two_factor_auth_enabled",
            "user.updated_at",
            "user.url",
            "user.verified",
            "user.iana_time_zone",
            "user.shared_phone_number",
            "user.role_type",
            "user.default_group_id",
            "user.report_csv",
            "target.active",
            "target.content_type",
            "target.created_at",
            "target.id",
            "target.method",
            "target.password",
            "target.target_url",
            "target.title",
            "target.type",
            "target.url",
            "target.username",
            "target.ticket.assignee_id",
            "target.ticket.brand_id",
            "target.ticket.collaborator_ids",
            "target.ticket.created_at",
            "target.ticket.custom_fields",
            "target.ticket.description",
            "target.ticket.due_at",
            "target.ticket.external_id",
            "target.ticket.fields",
            "target.ticket.forum_topic_id",
            "target.ticket.group_id",
            "target.ticket.has_incidents",
            "target.ticket.id",
            "target.ticket.organization_id",
            "target.ticket.priority",
            "target.ticket.problem_id",
            "target.ticket.raw_subject",
            "target.ticket.recipient",
            "target.ticket.requester_id",
            "target.ticket.satisfaction_rating",
            "target.ticket.sharing_agreement_ids",
            "target.ticket.status",
            "target.ticket.subject",
            "target.ticket.submitter_id",
            "target.ticket.tags",
            "target.ticket.type",
            "target.ticket.updated_at",
            "target.ticket.url",
            "target.ticket.via"
        ]



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE

connection_args = OrderedDict(
    bucket={
        'type': ARG_TYPE.STR,
        'description': 'The name of the GCS bucket.',
        'label': 'GCS Bucket'
    },
    service_account_keys={
        'type': ARG_TYPE.PATH,
        'description': 'Path to the service account JSON file',
        'label': 'Path to the service account JSON file',
        'secret': True
    },
    service_account_json={
        'type': ARG_TYPE.DICT,
        'description': 'Content of service account JSON file',
        'label': 'Content of service account JSON file',
        'secret': True
    }
)


connection_args_example = OrderedDict(
    bucket='my-bucket',
    service_account_keys='/Users/sam/Downloads/svc.json'
)



from dataclasses import dataclass, field
from typing import Optional, List, Dict

import pandas as pd


@dataclass
class DataHubResponse:
    data_frame: pd.DataFrame = field(default_factory=pd.DataFrame)
    columns: List[Dict] = field(default_factory=list)
    affected_rows: Optional[int] = None



import re

from bs4 import BeautifulSoup
import bs4.element
import chardet

import pandas as pd

from mindsdb.integrations.handlers.email_handler.email_client import EmailClient
from mindsdb.integrations.handlers.email_handler.settings import EmailSearchOptions


class EmailIngestor:
    """
    Parses emails into a DataFrame.
    Does some preprocessing on the raw HTML to extract meaningful text.
    """
    def __init__(self, email_client: EmailClient, search_options: EmailSearchOptions):
        self.email_client = email_client
        self.search_options = search_options

    def _is_tag_visible(self, element):
        if element.parent.name in ['style', 'script', 'head', 'title', 'meta', '[document]']:
            return False
        if isinstance(element, bs4.element.Comment):
            return False
        return True

    def _preprocess_raw_html(self, html: str) -> str:
        soup = BeautifulSoup(html, 'html.parser')
        texts = soup.find_all(text=True)
        visible_texts = filter(self._is_tag_visible, texts)
        return '\n'.join(t.strip() for t in visible_texts)

    def _ingest_email_row(self, row: pd.Series) -> dict:
        if row['body_content_type'] == 'html':
            # Extract meaningful text from raw HTML.
            row['body'] = self._preprocess_raw_html(row['body'])
        body_str = row['body']
        encoding = None
        if isinstance(body_str, bytes):
            encoding = chardet.detect(body_str)['encoding']
            if 'windows' in encoding.lower():
                # Easier to treat this at utf-8 since str constructor doesn't support all encodings here:
                # https://chardet.readthedocs.io/en/latest/supported-encodings.html.
                encoding = 'utf-8'
            try:
                body_str = str(body_str, encoding=encoding)
            except UnicodeDecodeError:
                # If illegal characters are found, we ignore them.
                # I encountered this issue with some emails that had a mix of encodings.
                body_str = row['body'].decode(encoding, errors='ignore')
        # We split by paragraph so make sure there aren't too many newlines in a row.
        body_str = re.sub(r'[\r\n]\s*[\r\n]', '\n\n', body_str)
        email_data = {
            'id': row['id'],
            'body': body_str,
            'subject': row['subject'],
            'to_field': row['to_field'],
            'from_field': row['from_field'],
            'datetime': row['date']
        }
        # Replacing None values {None: ""}
        for key in email_data:
            if email_data[key] is None:
                email_data[key] = ""

        return email_data

    def ingest(self) -> pd.DataFrame:
        emails_df = self.email_client.search_email(self.search_options)
        all_email_data = []
        for _, row in emails_df.iterrows():
            all_email_data.append(self._ingest_email_row(row))

        df = pd.DataFrame(all_email_data)

        # Replace "(UTC)" with empty string over a pandas DataFrame column
        if 'datetime' in df.columns:
            df['datetime'] = df['datetime'].str.replace(' (UTC)', '')

            # Convert datetime string to datetime object, and normalize timezone to UTC.
            df['datetime'] = pd.to_datetime(df['datetime'], utc=True, format="%a, %d %b %Y %H:%M:%S %z", errors='coerce')

        return df




"""agents-deleted-at

Revision ID: 8e17ff6b75e9
Revises: 45eb2eb61f70
Create Date: 2024-08-12 19:13:44.327111

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db  # noqa


# revision identifiers, used by Alembic.
revision = '8e17ff6b75e9'
down_revision = '45eb2eb61f70'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('agents', schema=None) as batch_op:
        batch_op.add_column(sa.Column('deleted_at', sa.DateTime(), nullable=True))

    with op.batch_alter_table('skills', schema=None) as batch_op:
        batch_op.add_column(sa.Column('created_at', sa.DateTime(), nullable=True))
        batch_op.add_column(sa.Column('updated_at', sa.DateTime(), nullable=True))
        batch_op.add_column(sa.Column('deleted_at', sa.DateTime(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('skills', schema=None) as batch_op:
        batch_op.drop_column('deleted_at')
        batch_op.drop_column('updated_at')
        batch_op.drop_column('created_at')

    with op.batch_alter_table('agents', schema=None) as batch_op:
        batch_op.drop_column('deleted_at')
    # ### end Alembic commands ###



""" BYOM: Bring Your Own Model

env vars to contloll BYOM:
 - MINDSDB_BYOM_ENABLED - can BYOM be uysed or not. Locally enabled by default.
 - MINDSDB_BYOM_INHOUSE_ENABLED - enable or disable 'inhouse' BYOM usage. Locally enabled by default.
 - MINDSDB_BYOM_DEFAULT_TYPE - [inhouse|venv] default byom type. Locally it is 'venv' by default.
 - MINDSDB_BYOM_TYPE - [safe|unsafe] - obsolete, same as above.
"""


import os
import re
import sys
import shutil
import pickle
import tarfile
import tempfile
import traceback
import subprocess
from enum import Enum
from pathlib import Path
from datetime import datetime
from typing import Optional, Dict, Union

import pandas as pd
from pandas.api import types as pd_types

from mindsdb.utilities import log
from mindsdb.utilities.config import Config
from mindsdb.utilities.fs import safe_extract
from mindsdb.interfaces.storage import db
from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.integrations.libs.const import PREDICTOR_STATUS
from mindsdb.integrations.utilities.utils import format_exception_error
import mindsdb.utilities.profiler as profiler


from .proc_wrapper import (
    pd_decode, pd_encode, encode, decode, BYOM_METHOD,
    import_string, find_model_class, check_module
)
from .__about__ import __version__


BYOM_TYPE = Enum('BYOM_TYPE', ['INHOUSE', 'VENV'])

logger = log.getLogger(__name__)


class BYOMHandler(BaseMLEngine):

    name = 'byom'

    def __init__(self, model_storage, engine_storage, **kwargs) -> None:
        # region check availability
        is_cloud = Config().get('cloud', False)
        if is_cloud is True:
            byom_enabled = os.environ.get('MINDSDB_BYOM_ENABLED', 'false').lower()
            if byom_enabled not in ('true', '1'):
                raise RuntimeError('BYOM is disabled on cloud')
        # endregion

        self.model_wrapper = None

        self.inhouse_model_wrapper = None
        self.model_wrappers = {}

        # region read and save set default byom type
        try:
            self._default_byom_type = BYOM_TYPE.VENV
            if os.environ.get('MINDSDB_BYOM_DEFAULT_TYPE') is not None:
                self._default_byom_type = BYOM_TYPE[
                    os.environ.get('MINDSDB_BYOM_DEFAULT_TYPE').upper()
                ]
            else:
                env_var = os.environ.get('MINDSDB_BYOM_DEFAULT_TYPE')
                if env_var == 'SAVE':
                    self._default_byom_type = BYOM_TYPE['VENV']
                elif env_var == 'UNSAVE':
                    self._default_byom_type = BYOM_TYPE['INHOUSE']
                else:
                    raise KeyError
        except KeyError:
            self._default_byom_type = BYOM_TYPE.VENV
        # endregion

        # region check if 'inhouse' BYOM is enabled
        env_var = os.environ.get('MINDSDB_BYOM_INHOUSE_ENABLED')
        if env_var is None:
            self._inhouse_enabled = False if is_cloud else True
        else:
            self._inhouse_enabled = env_var.lower() in ('true', '1')
        # endregion

        super().__init__(model_storage, engine_storage, **kwargs)

    @staticmethod
    def normalize_engine_version(engine_version: Union[int, str, None]) -> int:
        """Cast engine version to int, or return `1` if can not be casted

        Args:
            engine_version (Union[int, str, None]): engine version

        Returns:
            int: engine version
        """
        if isinstance(engine_version, str):
            try:
                engine_version = int(engine_version)
            except Exception:
                engine_version = 1
        if isinstance(engine_version, int) is False:
            engine_version = 1
        return engine_version

    @staticmethod
    def create_validation(target: str, args: dict = None, **kwargs) -> None:
        if isinstance(args, dict) is False:
            return
        using_args = args.get('using', {})
        engine_version = using_args.get('engine_version')
        if engine_version is not None:
            engine_version = BYOMHandler.normalize_engine_version(engine_version)
        else:
            connection_args = kwargs['handler_storage'].get_connection_args()
            versions = connection_args.get('versions')
            if isinstance(versions, dict):
                engine_version = max([int(x) for x in versions.keys()])
            else:
                engine_version = 1
            using_args['engine_version'] = engine_version

    def get_model_engine_version(self) -> int:
        """Return current model engine version

        Returns:
            int: engine version
        """
        engine_version = self.model_storage.get_info()['learn_args'].get('using', {}).get('engine_version')
        engine_version = BYOMHandler.normalize_engine_version(engine_version)
        return engine_version

    def normalize_byom_type(self, byom_type: Optional[str]) -> BYOM_TYPE:
        if byom_type is not None:
            byom_type = BYOM_TYPE[byom_type.upper()]
        else:
            byom_type = self._default_byom_type
        if byom_type == BYOM_TYPE.INHOUSE and self._inhouse_enabled is False:
            raise Exception("'Inhouse' BYOM engine type can not be used")
        return byom_type

    def _get_model_proxy(self, version=None):
        if version is None:
            version = 1
        if isinstance(version, str):
            version = int(version)
        version_mark = ''
        if version > 1:
            version_mark = f'_{version}'
        version_str = str(version)

        self.engine_storage.fileStorage.pull()
        try:
            code = self.engine_storage.fileStorage.file_get(f'code{version_mark}')
            modules_str = self.engine_storage.fileStorage.file_get(f'modules{version_mark}')
        except FileNotFoundError:
            raise Exception(f"Engine version '{version}' does not exists")

        if version_str not in self.model_wrappers:
            connection_args = self.engine_storage.get_connection_args()
            version_meta = connection_args['versions'][version_str]

            try:
                engine_version_type = BYOM_TYPE[
                    version_meta.get('type', self._default_byom_type.name).upper()
                ]
            except KeyError:
                raise Exception('Unknown BYOM engine type')

            if engine_version_type == BYOM_TYPE.INHOUSE:
                if self._inhouse_enabled is False:
                    raise Exception("'Inhouse' BYOM engine type can not be used")
                if self.inhouse_model_wrapper is None:
                    self.inhouse_model_wrapper = ModelWrapperUnsafe(
                        code=code,
                        modules_str=modules_str,
                        engine_id=self.engine_storage.integration_id,
                        engine_version=version
                    )
                self.model_wrappers[version_str] = self.inhouse_model_wrapper
            elif engine_version_type == BYOM_TYPE.VENV:
                if version_meta.get('venv_status') != 'ready':
                    version_meta['venv_status'] = 'creating'
                    self.engine_storage.update_connection_args(connection_args)
                self.model_wrappers[version_str] = ModelWrapperSafe(
                    code=code,
                    modules_str=modules_str,
                    engine_id=self.engine_storage.integration_id,
                    engine_version=version
                )
                version_meta['venv_status'] = 'ready'
                self.engine_storage.update_connection_args(connection_args)

        return self.model_wrappers[version_str]

    def describe(self, attribute: Optional[str] = None) -> pd.DataFrame:
        engine_version = self.get_model_engine_version()
        mp = self._get_model_proxy(engine_version)
        model_state = self.model_storage.file_get('model')
        return mp.describe(model_state, attribute)

    def create(self, target, df=None, args=None, **kwargs):
        using_args = args.get('using', {})
        engine_version = using_args.get('engine_version')

        model_proxy = self._get_model_proxy(engine_version)
        model_state = model_proxy.train(df, target, args)

        self.model_storage.file_set('model', model_state)

        # TODO return columns?

        def convert_type(field_type):
            if pd_types.is_integer_dtype(field_type):
                return 'integer'
            elif pd_types.is_numeric_dtype(field_type):
                return 'float'
            elif pd_types.is_datetime64_any_dtype(field_type):
                return 'datetime'
            else:
                return 'categorical'

        columns = {
            target: convert_type(object)
        }

        self.model_storage.columns_set(columns)

    def predict(self, df, args=None):
        pred_args = args.get('predict_params', {})

        engine_version = pred_args.get('engine_version')
        if engine_version is not None:
            engine_version = int(engine_version)
        else:
            engine_version = self.get_model_engine_version()

        model_proxy = self._get_model_proxy(engine_version)
        model_state = self.model_storage.file_get('model')
        pred_df = model_proxy.predict(df, model_state, pred_args)

        return pred_df

    def create_engine(self, connection_args):
        code_path = Path(connection_args['code'])
        self.engine_storage.fileStorage.file_set(
            'code',
            code_path.read_bytes()
        )

        requirements_path = Path(connection_args['modules'])
        self.engine_storage.fileStorage.file_set(
            'modules',
            requirements_path.read_bytes()
        )

        self.engine_storage.fileStorage.push()

        self.engine_storage.update_connection_args({
            'handler_version': __version__,
            'mode': connection_args.get('mode'),
            'versions': {
                '1': {
                    'code': code_path.name,
                    'requirements': requirements_path.name,
                    'type': self.normalize_byom_type(
                        connection_args.get('type')
                    ).name.lower()
                }
            }
        })

        model_proxy = self._get_model_proxy()
        try:
            info = model_proxy.check(connection_args.get('mode'))
            self.engine_storage.json_set('methods', info['methods'])

        except Exception as e:
            if hasattr(model_proxy, 'remove_venv'):
                model_proxy.remove_venv()
            raise e

    def update_engine(self, connection_args: dict) -> None:
        """Add new version of engine

            Args:
                connection_args (dict): paths to code and requirements
        """
        code_path = Path(connection_args['code'])
        requirements_path = Path(connection_args['modules'])

        engine_connection_args = self.engine_storage.get_connection_args()
        if isinstance(engine_connection_args, dict) is False or 'handler_version' not in engine_connection_args:
            engine_connection_args = {
                'handler_version': __version__,
                'versions': {
                    '1': {
                        'code': 'code.py',
                        'requirements': 'requirements.txt',
                        'type': self._default_byom_type.name.lower()
                    }
                }
            }
        new_version = str(max([int(x) for x in engine_connection_args['versions'].keys()]) + 1)

        engine_connection_args['versions'][new_version] = {
            'code': code_path.name,
            'requirements': requirements_path.name,
            'type': self.normalize_byom_type(
                connection_args.get('type')
            ).name.lower()
        }

        self.engine_storage.fileStorage.file_set(
            f'code_{new_version}',
            code_path.read_bytes()
        )

        self.engine_storage.fileStorage.file_set(
            f'modules_{new_version}',
            requirements_path.read_bytes()
        )
        self.engine_storage.fileStorage.push()

        self.engine_storage.update_connection_args(engine_connection_args)

        model_proxy = self._get_model_proxy(new_version)
        try:
            methods = model_proxy.check()
            self.engine_storage.json_set('methods', methods)

        except Exception as e:
            if hasattr(model_proxy, 'remove_venv'):
                model_proxy.remove_venv()
            raise e

    def function_list(self):
        return self.engine_storage.json_get('methods')

    def function_call(self, name, args):
        mp = self._get_model_proxy()
        return mp.func_call(name, args)

    def finetune(self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        using_args = args.get('using', {})
        engine_version = using_args.get('engine_version')

        model_storage = self.model_storage
        # TODO: should probably refactor at some point, as a bit of the logic is shared with lightwood's finetune logic
        try:
            base_predictor_id = args['base_model_id']
            base_predictor_record = db.Predictor.query.get(base_predictor_id)
            if base_predictor_record.status != PREDICTOR_STATUS.COMPLETE:
                raise Exception("Base model must be in status 'complete'")

            predictor_id = model_storage.predictor_id
            predictor_record = db.Predictor.query.get(predictor_id)

            predictor_record.data = {'training_log': 'training'}  # TODO move to ModelStorage (don't work w/ db directly)
            predictor_record.training_start_at = datetime.now()
            predictor_record.status = PREDICTOR_STATUS.FINETUNING  # TODO: parallel execution block
            db.session.commit()

            model_proxy = self._get_model_proxy(engine_version)
            model_state = self.base_model_storage.file_get('model')
            model_state = model_proxy.finetune(df, model_state, args=args.get('using', {}))

            # region hack to speedup file saving
            with profiler.Context('finetune-byom-write-file'):
                dest_abs_path = model_storage.fileStorage.folder_path / 'model'
                with open(dest_abs_path, 'wb') as fd:
                    fd.write(model_state)
                model_storage.fileStorage.push(compression_level=0)
            # endregion

            predictor_record.update_status = 'up_to_date'
            predictor_record.status = PREDICTOR_STATUS.COMPLETE
            predictor_record.training_stop_at = datetime.now()
            db.session.commit()

        except Exception as e:
            logger.error(e)
            predictor_id = model_storage.predictor_id
            predictor_record = db.Predictor.query.with_for_update().get(predictor_id)
            logger.error(traceback.format_exc())
            error_message = format_exception_error(e)
            predictor_record.data = {"error": error_message}
            predictor_record.status = PREDICTOR_STATUS.ERROR
            db.session.commit()
            raise

        finally:
            if predictor_record.training_stop_at is None:
                predictor_record.training_stop_at = datetime.now()
                db.session.commit()


class ModelWrapperUnsafe:
    """ Model wrapper that executes learn/predict in current process
    """

    def __init__(self, code, modules_str, engine_id, engine_version: int):
        self.module = import_string(code)

        model_instance = None
        model_class = find_model_class(self.module)
        if model_class is not None:
            model_instance = model_class()

        self.model_instance = model_instance

    def train(self, df, target, args):
        self.model_instance.train(df, target, args)
        return pickle.dumps(self.model_instance.__dict__, protocol=5)

    def predict(self, df, model_state, args):
        model_state = pickle.loads(model_state)
        self.model_instance.__dict__ = model_state
        try:
            result = self.model_instance.predict(df, args)
        except Exception:
            result = self.model_instance.predict(df)
        return result

    def finetune(self, df, model_state, args):
        self.model_instance.__dict__ = pickle.loads(model_state)

        call_args = [df]
        if args:
            call_args.append(args)

        self.model_instance.finetune(df, args)

        return pickle.dumps(self.model_instance.__dict__, protocol=5)

    def describe(self, model_state, attribute: Optional[str] = None) -> pd.DataFrame:
        if hasattr(self.model_instance, 'describe'):
            model_state = pickle.loads(model_state)
            self.model_instance.__dict__ = model_state
            return self.model_instance.describe(attribute)
        return pd.DataFrame()

    def func_call(self, func_name, args):
        func = getattr(self.module, func_name)
        return func(*args)

    def check(self, mode: str = None):
        methods = check_module(self.module, mode)
        return methods


class ModelWrapperSafe:
    """ Model wrapper that executes learn/predict in venv
    """

    def __init__(self, code, modules_str, engine_id, engine_version: int):
        self.code = code
        modules = self.parse_requirements(modules_str)

        self.config = Config()
        self.is_cloud = Config().get('cloud', False)

        self.env_path = None
        self.env_storage_path = None
        self.prepare_env(modules, engine_id, engine_version)

    def prepare_env(self, modules, engine_id, engine_version: int):
        try:
            import virtualenv

            base_path = self.config.get('byom', {}).get('venv_path')
            if base_path is None:
                # create in root path
                base_path = Path(self.config.paths['root']) / 'venvs'
            else:
                base_path = Path(base_path)
            base_path.mkdir(parents=True, exist_ok=True)

            env_folder_name = f'env_{engine_id}'
            if isinstance(engine_version, int) and engine_version > 1:
                env_folder_name = f'{env_folder_name}_{engine_version}'

            self.env_storage_path = base_path / env_folder_name
            if self.is_cloud:
                bese_env_path = Path(tempfile.gettempdir()) / 'mindsdb' / 'venv'
                bese_env_path.mkdir(parents=True, exist_ok=True)
                self.env_path = bese_env_path / env_folder_name
                tar_path = self.env_storage_path.with_suffix('.tar')
                if self.env_path.exists() is False and tar_path.exists() is True:
                    with tarfile.open(tar_path) as tar:
                        safe_extract(tar, path=bese_env_path)
            else:
                self.env_path = self.env_storage_path

            if sys.platform in ('win32', 'cygwin'):
                exectable_folder_name = 'Scripts'
            else:
                exectable_folder_name = 'bin'

            pip_cmd = self.env_path / exectable_folder_name / 'pip'
            self.python_path = self.env_path / exectable_folder_name / 'python'

            if self.env_path.exists():
                # already exists. it means requirements are already installed
                return

            # create
            logger.info(f"Creating new environment: {self.env_path}")
            virtualenv.cli_run(['-p', sys.executable, str(self.env_path)])
            logger.info(f"Created new environment: {self.env_path}")

            if len(modules) > 0:
                self.install_modules(modules, pip_cmd=pip_cmd)
        except Exception:
            # DANGER !!! VENV MUST BE CREATED
            logger.info("Can't create virtual environment. venv module should be installed")

            if self.is_cloud:
                raise

            self.python_path = Path(sys.executable)

            # try to install modules everytime
            self.install_modules(modules, pip_cmd=pip_cmd)

        # fastest way to copy files if destination is NFS
        if self.is_cloud and self.env_storage_path != self.env_path:
            old_cwd = os.getcwd()
            os.chdir(str(bese_env_path))
            tar_path = self.env_path.with_suffix('.tar')
            with tarfile.open(name=str(tar_path), mode='w') as tar:
                tar.add(str(self.env_path.name))
            os.chdir(old_cwd)
            subprocess.run(
                ['cp', '-R', '--no-preserve=mode,ownership', str(tar_path), str(base_path / tar_path.name)],
                check=True, shell=False
            )
            tar_path.unlink()

    def remove_venv(self):
        if self.env_path is not None and self.env_path.exists():
            shutil.rmtree(str(self.env_path))

        if self.is_cloud:
            tar_path = self.env_storage_path.with_suffix('.tar')
            tar_path.unlink()

    def parse_requirements(self, requirements):
        # get requirements from string
        # they should be located at the top of the file, before code

        pattern = '^[\w\\[\\]-]+[=!<>\s]*[\d\.]*[,=!<>\s]*[\d\.]*$'  # noqa
        modules = []
        for line in requirements.split(b'\n'):
            line = line.decode().strip()
            if line:
                if re.match(pattern, line):
                    modules.append(line)
                else:
                    raise Exception(f'Wrong requirement: {line}')

        is_pandas = any([m.lower().startswith('pandas') for m in modules])
        if not is_pandas:
            modules.append('pandas>=2.0.0,<2.1.0')
            modules.append('numpy<2.0.0')

        # for dataframe serialization
        modules.append('pyarrow==19.0.0')
        return modules

    def install_modules(self, modules, pip_cmd):
        # install in current environment using pip
        for module in modules:
            logger.debug(f"BYOM install module: {module}")
            p = subprocess.Popen([pip_cmd, 'install', module], stderr=subprocess.PIPE)
            p.wait()
            if p.returncode != 0:
                raise Exception(f'Problem with installing module {module}: {p.stderr.read()}')

    def _run_command(self, params):
        logger.debug(f"BYOM run command: {params.get('method')}")
        params_enc = encode(params)

        wrapper_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'proc_wrapper.py')
        p = subprocess.Popen(
            [str(self.python_path), wrapper_path],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        p.stdin.write(params_enc)
        p.stdin.close()
        ret_enc = p.stdout.read()

        p.wait()

        try:
            ret = decode(ret_enc)
        except (pickle.UnpicklingError, EOFError):
            raise RuntimeError(p.stderr.read())
        return ret

    def check(self, mode: str = None):
        params = {
            'method': BYOM_METHOD.CHECK.value,
            'code': self.code,
            'mode': mode,
        }
        return self._run_command(params)

    def train(self, df, target, args):
        params = {
            'method': BYOM_METHOD.TRAIN.value,
            'code': self.code,
            'df': None,
            'to_predict': target,
            'args': args,
        }
        if df is not None:
            params['df'] = pd_encode(df)

        model_state = self._run_command(params)
        return model_state

    def predict(self, df, model_state, args):
        params = {
            'method': BYOM_METHOD.PREDICT.value,
            'code': self.code,
            'model_state': model_state,
            'df': pd_encode(df),
            'args': args,
        }
        pred_df = self._run_command(params)
        return pd_decode(pred_df)

    def finetune(self, df, model_state, args):
        params = {
            'method': BYOM_METHOD.FINETUNE.value,
            'code': self.code,
            'model_state': model_state,
            'df': pd_encode(df),
            'args': args,
        }

        model_state = self._run_command(params)
        return model_state

    def describe(self, model_state, attribute: Optional[str] = None) -> pd.DataFrame:
        params = {
            'method': BYOM_METHOD.DESCRIBE.value,
            'code': self.code,
            'model_state': model_state,
            'attribute': attribute
        }
        enc_df = self._run_command(params)
        df = pd_decode(enc_df)
        return df

    def func_call(self, func_name, args):
        params = {
            'method': BYOM_METHOD.FUNC_CALL.value,
            'code': self.code,
            'func_name': func_name,
            'args': args,
        }
        result = self._run_command(params)
        return result



from typing import Optional

from mindsdb_sql_parser.ast.base import ASTNode

from mindsdb.utilities import log
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse
)

from mindsdb.integrations.handlers.mysql_handler.mysql_handler import MySQLHandler
from mindsdb.integrations.handlers.postgres_handler.postgres_handler import PostgresHandler
from mindsdb.integrations.handlers.mssql_handler.mssql_handler import SqlServerHandler


class CloudSQLHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Google Cloud SQL statements.
    """
    name = 'cloud_sql'

    def __init__(self, name: str, connection_data: Optional[dict], **kwargs):
        """
        Initialize the handler.
        Args:
            name (str): name of particular handler instance
            connection_data (dict): parameters for connecting to the database
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)

        self.dialect = 'cloud_sql'
        self.connection_data = connection_data
        self.kwargs = kwargs

        if self.connection_data['db_engine'] == 'mysql':
            self.db = MySQLHandler(
                name=name + 'mysql',
                connection_data={key: self.connection_data[key] for key in self.connection_data if key != 'db_engine'}
            )
        elif self.connection_data['db_engine'] == 'postgresql':
            self.db = PostgresHandler(
                name=name + 'postgresql',
                connection_data={key: self.connection_data[key] for key in self.connection_data if key != 'db_engine'}
            )
        elif self.connection_data['db_engine'] == 'mssql':
            self.db = SqlServerHandler(
                name=name + 'mssql',
                connection_data={key: self.connection_data[key] for key in self.connection_data if key != 'db_engine'}
            )
        else:
            raise Exception("The database engine should be either MySQL, PostgreSQL or SQL Server!")

    def __del__(self):
        self.db.__del__()

    def connect(self) -> StatusResponse:
        """
        Set up the connection required by the handler.
        Returns:
            HandlerStatusResponse
        """

        return self.db.connect()

    def disconnect(self):
        """
        Close any existing connections.
        """

        return self.db.disconnect()

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler.
        Returns:
            HandlerStatusResponse
        """

        return self.db.check_connection()

    def native_query(self, query: str) -> StatusResponse:
        """
        Receive raw query and act upon it somehow.
        Args:
            query (str): query in native format
        Returns:
            HandlerResponse
        """

        return self.db.native_query(query)

    def query(self, query: ASTNode) -> StatusResponse:
        """
        Receive query as AST (abstract syntax tree) and act upon it somehow.
        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INTSERT, DELETE, etc
        Returns:
            HandlerResponse
        """

        return self.db.query(query)

    def get_tables(self) -> StatusResponse:
        """
        Return list of entities that will be accessible as tables.
        Returns:
            HandlerResponse
        """

        return self.db.get_tables()

    def get_columns(self, table_name: str) -> StatusResponse:
        """
        Returns a list of entity columns.
        Args:
            table_name (str): name of one of tables returned by self.get_tables()
        Returns:
            HandlerResponse
        """

        return self.db.get_columns(table_name)



import github

from mindsdb_sql_parser import parse_sql

from mindsdb.integrations.handlers.github_handler.github_tables import (
    GithubIssuesTable,
    GithubPullRequestsTable,
    GithubCommitsTable,
    GithubReleasesTable,
    GithubBranchesTable,
    GithubContributorsTable,
    GithubMilestonesTable,
    GithubProjectsTable, GithubFilesTable
)
from mindsdb.integrations.handlers.github_handler.generate_api import get_github_types, get_github_methods, GHTable
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)
from mindsdb.utilities import log


logger = log.getLogger(__name__)


class GithubHandler(APIHandler):
    """The GitHub handler implementation"""

    def __init__(self, name: str, **kwargs):
        """Initialize the GitHub handler.

        Parameters
        ----------
        name : str
            name of a handler instance
        """
        super().__init__(name)

        connection_data = kwargs.get("connection_data", {})
        self.connection_data = connection_data
        self.repository = connection_data["repository"]
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

        # custom tables
        self._register_table("issues", GithubIssuesTable(self))
        self._register_table("files", GithubFilesTable(self))

        # generated tables
        github_types = get_github_types()

        # generate tables from repository object
        for method in get_github_methods(github.Repository.Repository):
            if method.table_name in self._tables:
                continue

            table = GHTable(self, github_types=github_types, method=method)
            self._register_table(method.table_name, table)


    def connect(self) -> StatusResponse:
        """Set up the connection required by the handler.

        Returns
        -------
        StatusResponse
            connection object
        """

        if self.is_connected is True:
            return self.connection

        connection_kwargs = {}

        if self.connection_data.get("api_key", None):
            connection_kwargs["login_or_token"] = self.connection_data["api_key"]

        if self.connection_data.get("github_url", None):
            connection_kwargs["base_url"] = self.connection_data["github_url"]

        self.connection = github.Github(**connection_kwargs)
        self.is_connected = True

        return self.connection

    def check_connection(self) -> StatusResponse:
        """Check connection to the handler.

        Returns
        -------
        StatusResponse
            Status confirmation
        """
        response = StatusResponse(False)

        try:
            self.connect()
            if self.connection_data.get("api_key", None):
                current_user = self.connection.get_user().name
                logger.info(f"Authenticated as user {current_user}")
            else:
                logger.info("Proceeding without an API key")

            current_limit = self.connection.get_rate_limit()
            logger.info(f"Current rate limit: {current_limit}")
            response.success = True
        except Exception as e:
            logger.error(f"Error connecting to GitHub API: {e}!")
            response.error_message = e

        self.is_connected = response.success

        return response

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query.

        Parameters
        ----------
        query : str
            query in a native format

        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query)
        return self.query(ast)



import os
from pyzotero import zotero
from mindsdb.utilities import log
from mindsdb.utilities.config import Config
from mindsdb.integrations.libs.api_handler import APIHandler, FuncParser

from mindsdb.integrations.handlers.zotero_handler.zotero_tables import AnnotationsTable

from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)

logger = log.getLogger(__name__)


class ZoteroHandler(APIHandler):
    """Handles communication with the Zotero API."""

    def __init__(self, name=None, **kwargs):
        """Initialize the Zotero handler.

        Parameters
        ----------
        name : str
            Name of the handler instance.

        Other Parameters
        ----------------
        connection_data : dict
            Dictionary containing connection data such as 'library_id', 'library_type', and 'api_key'.
            If not provided, will attempt to fetch from environment variables or configuration file.
        """
        super().__init__(name)
        self.connection_args = self._get_connection_args(kwargs.get('connection_data', {}))
        self.is_connected = False
        self.api = None
        self._register_table('annotations', AnnotationsTable(self))

    def _get_connection_args(self, args):
        """Fetch connection arguments from parameters, environment variables, or configuration.

        Parameters
        ----------
        args
            Dictionary containing connection data.

        Returns
        -------
        connection_args
            Connection data list
        """
        handler_config = Config().get('zotero_handler', {})
        connection_args = {}
        for k in ['library_id', 'library_type', 'api_key']:
            connection_args[k] = args.get(k) or os.getenv(f'ZOTERO_{k.upper()}') or handler_config.get(k)
        return connection_args

    def connect(self) -> StatusResponse:
        """Connect to the Zotero API.

        Returns
        -------
        StatusResponse
            Status of the connection attempt.
        """
        if not self.is_connected:
            self.api = zotero.Zotero(
                self.connection_args['library_id'],
                self.connection_args['library_type'],
                self.connection_args['api_key']
            )
            self.is_connected = True
        return StatusResponse(True)

    def check_connection(self) -> StatusResponse:
        """Check the connection status to the Zotero API.

        Returns
        -------
        StatusResponse
            Status of the connection.
        """
        try:
            self.connect()
            return StatusResponse(True)
        except Exception as e:
            error_message = f'Error connecting to Zotero API: {str(e)}. Check credentials.'
            logger.error(error_message)
            self.is_connected = False
            return StatusResponse(False, error_message=error_message)

    def native_query(self, query_string: str = None):
        """Execute a native query against the Zotero API.

        Parameters
        ----------
        query_string : str
            The query string to execute, formatted as required by the Zotero API.

        Returns
        -------
        Response
            Response object containing the result of the query.
        """
        method_name, params = FuncParser().from_string(query_string)
        df = self._call_find_annotations_zotero_api(method_name, params)
        return Response(RESPONSE_TYPE.TABLE, data_frame=df)



import os
import time
import signal
import tempfile
import threading
from pathlib import Path
from functools import wraps
from collections.abc import Callable

import psutil
from walrus import Database
from pandas import DataFrame
from redis.exceptions import ConnectionError as RedisConnectionError

from mindsdb.utilities.config import Config
from mindsdb.utilities.context import context as ctx
from mindsdb.integrations.libs.process_cache import process_cache
from mindsdb.utilities.ml_task_queue.utils import RedisKey, StatusNotifier, to_bytes, from_bytes
from mindsdb.utilities.ml_task_queue.base import BaseRedisQueue
from mindsdb.utilities.fs import clean_unlinked_process_marks
from mindsdb.utilities.functions import mark_process
from mindsdb.utilities.ml_task_queue.const import (
    ML_TASK_TYPE,
    ML_TASK_STATUS,
    TASKS_STREAM_NAME,
    TASKS_STREAM_CONSUMER_NAME,
    TASKS_STREAM_CONSUMER_GROUP_NAME
)
from mindsdb.utilities import log
from mindsdb.utilities.sentry import sentry_sdk  # noqa: F401

logger = log.getLogger(__name__)


def _save_thread_link(func: Callable) -> Callable:
    """ Decorator for MLTaskConsumer.
        Save thread in which func is executed to a list.
    """
    @wraps(func)
    def wrapper(self, *args, **kwargs) -> None:
        current_thread = threading.current_thread()
        self._listen_message_threads.append(current_thread)
        try:
            result = func(self, *args, **kwargs)
        finally:
            self._listen_message_threads.remove(current_thread)
        return result
    return wrapper


class MLTaskConsumer(BaseRedisQueue):
    """ Listener of ML tasks queue and tasks executioner.
        Each new message waited and executed in separate thread.

        Attributes:
            _ready_event (Event): set if ready to start new queue listen thread
            _stop_event (Event): set if need to stop all threads/processes
            cpu_stat (list[float]): CPU usage statistic. Each value is 0-100 float representing CPU usage in %
            _collect_cpu_stat_thread (Thread): pointer to thread that collecting CPU usage statistic
            _listen_message_threads (list[Thread]): list of pointers to threads where queue messages are listening/processing
            db (Redis): database object
            cache: redis cache abstrtaction
            consumer_group: redis consumer group object
    """

    def __init__(self) -> None:
        self._ready_event = threading.Event()
        self._ready_event.set()

        self._stop_event = threading.Event()
        self._stop_event.clear()

        process_cache.init()

        # region collect cpu usage statistic
        self.cpu_stat = [0] * 10
        self._collect_cpu_stat_thread = threading.Thread(
            target=self._collect_cpu_stat, name='MLTaskConsumer._collect_cpu_stat'
        )
        self._collect_cpu_stat_thread.start()
        # endregion

        self._listen_message_threads = []

        # region connect to redis
        config = Config().get('ml_task_queue', {})
        self.db = Database(
            host=config.get('host', 'localhost'),
            port=config.get('port', 6379),
            db=config.get('db', 0),
            username=config.get('username'),
            password=config.get('password'),
            protocol=3
        )
        self.wait_redis_ping(60)

        self.db.Stream(TASKS_STREAM_NAME)
        self.cache = self.db.cache()
        self.consumer_group = self.db.consumer_group(TASKS_STREAM_CONSUMER_GROUP_NAME, [TASKS_STREAM_NAME])
        self.consumer_group.create()
        self.consumer_group.consumer(TASKS_STREAM_CONSUMER_NAME)
        # endregion

    def _collect_cpu_stat(self) -> None:
        """ Collect CPU usage statistic. Executerd in thread.
        """
        while self._stop_event.is_set() is False:
            self.cpu_stat = self.cpu_stat[1:]
            self.cpu_stat.append(psutil.cpu_percent())
            time.sleep(1)

    def get_avg_cpu_usage(self) -> float:
        """ get average CPU usage for last period (10s by default)

            Returns:
                float: 0-100 value, average CPU usage
        """
        return sum(self.cpu_stat) / len(self.cpu_stat)

    def wait_free_resources(self) -> None:
        """ Sleep in thread untill there are free resources. Checks:
            - avg CPU usage is less than 60%
            - current CPU usage is less than 60%
            - current tasks count is less than (N CPU cores) / 8
        """
        config = Config()
        is_cloud = config.get('cloud', False)
        processes_dir = Path(tempfile.gettempdir()).joinpath('mindsdb/processes/learn/')
        while True:
            while self.get_avg_cpu_usage() > 60 or max(self.cpu_stat[-3:]) > 60:
                time.sleep(1)
            if is_cloud and processes_dir.is_dir():
                clean_unlinked_process_marks()
                while (len(list(processes_dir.iterdir())) * 8) >= os.cpu_count():
                    time.sleep(1)
                    clean_unlinked_process_marks()
            if (self.get_avg_cpu_usage() > 60 or max(self.cpu_stat[-3:]) > 60) is False:
                return

    @_save_thread_link
    def _listen(self) -> None:
        """ Listen message queue untill get new message. Execute task.
        """
        message = None
        while message is None:
            self.wait_free_resources()
            self.wait_redis_ping()
            if self._stop_event.is_set():
                return

            try:
                message = self.consumer_group.read(count=1, block=1000, consumer=TASKS_STREAM_CONSUMER_NAME)
            except RedisConnectionError as e:
                logger.error(f"Can't connect to Redis: {e}")
                self._stop_event.set()
                return
            except Exception:
                self._stop_event.set()
                raise

            if message.get(TASKS_STREAM_NAME) is None or len(message.get(TASKS_STREAM_NAME)) == 0:
                message = None

        try:
            message = message[TASKS_STREAM_NAME][0][0]
            message_id = message[0].decode()
            message_content = message[1]
            self.consumer_group.streams[TASKS_STREAM_NAME].ack(message_id)
            self.consumer_group.streams[TASKS_STREAM_NAME].delete(message_id)

            payload = from_bytes(message_content[b'payload'])
            task_type = ML_TASK_TYPE(message_content[b'task_type'])
            model_id = int(message_content[b'model_id'])
            company_id = message_content[b'company_id']
            if len(company_id) == 0:
                company_id = None
            redis_key = RedisKey(message_content.get(b'redis_key'))

            # region read dataframe
            dataframe_bytes = self.cache.get(redis_key.dataframe)
            dataframe = None
            if dataframe_bytes is not None:
                dataframe = from_bytes(dataframe_bytes)
                self.cache.delete(redis_key.dataframe)
            # endregion

            ctx.load(payload['context'])
        finally:
            self._ready_event.set()

        try:
            task = process_cache.apply_async(
                task_type=task_type,
                model_id=model_id,
                payload=payload,
                dataframe=dataframe
            )
            status_notifier = StatusNotifier(redis_key, ML_TASK_STATUS.PROCESSING, self.db, self.cache)
            status_notifier.start()
            result = task.result()
        except Exception as e:
            self.wait_redis_ping()
            status_notifier.stop()
            exception_bytes = to_bytes(e)
            self.cache.set(redis_key.exception, exception_bytes, 10)
            self.db.publish(redis_key.status, ML_TASK_STATUS.ERROR.value)
            self.cache.set(redis_key.status, ML_TASK_STATUS.ERROR.value, 180)
        else:
            self.wait_redis_ping()
            status_notifier.stop()
            if isinstance(result, DataFrame):
                dataframe_bytes = to_bytes(result)
                self.cache.set(redis_key.dataframe, dataframe_bytes, 10)
            self.db.publish(redis_key.status, ML_TASK_STATUS.COMPLETE.value)
            self.cache.set(redis_key.status, ML_TASK_STATUS.COMPLETE.value, 180)

    def run(self) -> None:
        """ Start new listen thread each time when _ready_event is set
        """
        self._ready_event.set()
        while self._stop_event.is_set() is False:
            self._ready_event.wait(timeout=1)
            if self._ready_event.is_set() is False:
                continue
            self._ready_event.clear()
            threading.Thread(target=self._listen, name='MLTaskConsumer._listen').start()
        self.stop()

    def stop(self) -> None:
        """ Stop all executing threads
        """
        self._stop_event.set()
        for thread in (*self._listen_message_threads, self._collect_cpu_stat_thread):
            try:
                if thread.is_alive():
                    thread.join()
            except Exception:
                pass


@mark_process(name='internal', custom_mark='ml_task_consumer')
def start(verbose: bool) -> None:
    """ Create task queue consumer and start listen the queue
    """
    consumer = MLTaskConsumer()
    signal.signal(signal.SIGTERM, lambda _x, _y: consumer.stop())
    try:
        consumer.run()
    except Exception as e:
        consumer.stop()
        logger.error(f'Got exception: {e}', flush=True)
        raise
    finally:
        logger.info('Consumer process stopped', flush=True)



import os
from typing import Dict, Optional

import pandas as pd
from anthropic import AI_PROMPT, HUMAN_PROMPT, Anthropic

from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.utilities import log
from mindsdb.utilities.config import Config

from mindsdb.integrations.utilities.handler_utils import get_api_key

logger = log.getLogger(__name__)


class AnthropicHandler(BaseMLEngine):
    """
    Integration with the Anthropic LLM Python Library
    """

    name = "anthropic"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.default_chat_model = "claude-2.1"
        self.supported_chat_models = ["claude-instant-1.2", "claude-2.1", "claude-3-opus-20240229", "claude-3-sonnet-20240229"]
        self.default_max_tokens = 100
        self.generative = True
        self.connection = None

    def create(
        self,
        target: str,
        df: Optional[pd.DataFrame] = None,
        args: Optional[Dict] = None,
    ) -> None:

        if "using" not in args:
            raise Exception(
                "Anthropic engine requires a USING clause! Refer to its documentation for more details."
            )

        if "model" not in args["using"]:
            args["using"]["model"] = self.default_chat_model
        elif args["using"]["model"] not in self.supported_chat_models:
            raise Exception(
                f"Invalid chat model. Please use one of {self.supported_chat_models}"
            )

        if "max_tokens" not in args["using"]:
            args["using"]["max_tokens"] = self.default_max_tokens

        self.model_storage.json_set("args", args)

    def predict(
        self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None
    ) -> None:

        args = self.model_storage.json_get("args")
        api_key = get_api_key('anthropic', args["using"], self.engine_storage, strict=False)

        self.connection = Anthropic(
            api_key=api_key,
        )

        input_column = args["using"]["column"]

        if input_column not in df.columns:
            raise RuntimeError(f'Column "{input_column}" not found in input data')

        result_df = pd.DataFrame()

        result_df["predictions"] = df[input_column].apply(self.predict_answer)

        result_df = result_df.rename(columns={"predictions": args["target"]})

        return result_df

    def predict_answer(self, text):
        """
        connects with anthropic messages api to predict the answer for the particular question

        """

        args = self.model_storage.json_get("args")

        message = self.connection.messages.create(
            model=args["using"]["model"],
            max_tokens=args["using"]["max_tokens"],
            messages=[
                {"role": "user", "content": text}
            ]
        )

        content_blocks = message.content

        # assuming that message.content contains one ContentBlock item
        # returning text value if type==text and content_blocks value if type!=text
        if isinstance(content_blocks, list) and len(content_blocks) > 0:
            content_block = content_blocks[0]
            if content_block.type == 'text':
                return content_block.text
            else:
                return content_blocks
        else:
            raise Exception(
                f"Invalid output: {content_blocks}"
            )
